import numpy as np
import pandas as pd
from typing import Tuple, Optional, List, Union, Any, ClassVar, cast, Sequence, Literal
from dataclasses import dataclass, field
import warnings
from fanova import utils
from fanova.methods import oneway, twoway, plotting
import sys

@dataclass  # class to store these labels
class ANOVALabels:
    group: Optional[List[str]] = None
    primary: Optional[List[str]] = None
    secondary: Optional[List[str]] = None
    generic_group: Optional[bool] = True

    # Quantity Labels
    domain:Optional[str] = None     # such as Time
    response:Optional[str] = None   # Such as Temperature

    # Hypothesis Labels
    hypothesis: Optional[List[str]] = None
    H0_OneWay:ClassVar[Tuple[str, ...]] = ("FAMILY", "PAIRWISE")
    H0_TwoWay:ClassVar[Tuple[str, ...]] = ("FAMILY", "PAIRWISE", "INTERACTION", "PRIMARY", "SECONDARY", "CUSTOM")

@dataclass  # class to store units
class ANOVAUnits:
    domain: Optional[str] = None    # Such as seconds
    response: Optional[str] = None  # Such as Celcius

@dataclass
class ANOVATables:
    oneway: Optional[pd.DataFrame] = None
    oneway_bf: Optional[pd.DataFrame] = None
    twoway: Optional[pd.DataFrame] = None
    twoway_bf: Optional[pd.DataFrame] = None
    covar: Optional[pd.DataFrame] = None
    sw: Optional[pd.DataFrame] = None
    sig_figs = 4  # Significant Figures to display

@dataclass
class ANOVAMethods:  # clas to store methods
    # Class-level constants (shared by all instances, not settable per instance)
    anova_methods: ClassVar[Tuple[str, ...]] = (
        "L2-Simul", "L2-Naive", "L2-BiasReduced", "L2-Bootstrap",
        "F-Simul", "F-Naive", "F-BiasReduced", "F-Bootstrap"
    )

    covar_methods: ClassVar[Tuple[str, ...]] = (
        "L2-Simul", "L2-Naive", "L2-BiasReduced", "Permutation-Test", "Bootstrap-Test"
    )
    # Instance fields (can be changed internally if needed)
    anova_methods_used: Tuple[str, ...] = ()
    covar_methods_used: Tuple[str, ...] = ()

@dataclass
class ANOVAGroups:

    k: int = 0       # One-way ANOVA: number of groups
    A: int = 0       # Two-way ANOVA: primary levels
    B: int = 0       # Two-way ANOVA: secondary levels
    AB: int = 0      # Two-way ANOVA: total combinations

    subgroup_indicator: Union[np.ndarray, List[np.ndarray], None] = None # indicator Array for B

    contrast: Optional[Union[np.ndarray, List[int], None]] = None # User Specified Constrast vector
    contrast_factor:  Optional[int] = None # Either 1 for Primary, 2 for Secondary

@dataclass
class HypothesisInfo:  # Used just for vanilla Oneway
    SSH_t: np.ndarray
    pair_vec: list[str]
    q: int
    n_tests: int
    C: Optional[np.ndarray] = None
    D: Optional[np.ndarray] = None

@dataclass
class AnovaStatistics:  # Used just for vanilla Oneway
    T_n: np.ndarray
    F_n: Optional[np.ndarray] = None
    beta_hat: Optional[float] = None
    kappa_hat: Optional[float] = None
    beta_hat_unbias: Optional[float] = None
    kappa_hat_unbias: Optional[float] = None



class functionalANOVA():

    @property
    def tables(self) -> ANOVATables:
        return self._tables

    @property
    def methods(self) -> ANOVAMethods:
        return self._methods

    @property
    def labels(self) -> ANOVALabels:
        return  self._labels

    @property
    def units(self) -> ANOVAUnits:
        return  self._units

    @property
    def groups(self) -> ANOVAGroups:
        return  self._groups

    def __init__(
        self,
        data_list: List[np.ndarray] | Tuple[np.ndarray],
        d_grid: np.ndarray,
        grid_bounds: Tuple[float, float],
        n_boot: int = 10_000,
        n_simul: int = 10_000,
        alpha: float = 0.05,
        subgroup_indicator: Union[np.ndarray, List[np.ndarray], None] = None,
        group_labels: Optional[List[str]] = None,
        primary_labels: Optional[List[str]] = None,
        secondary_labels: Optional[List[str]] = None,
        domain_units: Optional[str] = None,
        domain_label: Optional[str] = None,
        response_units: Optional[str] = None,
        response_label: Optional[str] = None,
    ):


        self.data = []
        self.grid_bounds = grid_bounds
        self.d_grid = d_grid
        self.n_boot = int(n_boot)
        self.n_simul = int(n_simul)
        self.alpha = alpha
        self._labels = ANOVALabels(group_labels, primary_labels, secondary_labels)
        self._units = ANOVAUnits(domain_units, response_units)
        self._tables = ANOVATables()
        self._methods = ANOVAMethods()
        self._groups = ANOVAGroups(subgroup_indicator=subgroup_indicator)
        self.plottingOptions = plotting.PlotOptions()


        # Public and writable fields
        self.weights = "proportional"
        self.hypothesis = "FAMILY"
        self.verbose = True
        # self.show_simul_plots = False # Shows Null distribution plots for "Simul" Methods

        # Validate All Inputs
        self._validate_domain_response_labels(domain_units, domain_label,  response_units, response_label)
        self._validate_instantiation_inputs()

        self._groups.k = len(data_list)
        self.n_i = tuple(x.shape[1] for x in data_list)
        self.n_ii = None
        self.N = sum(self.n_i)    # Total Samples combined

        n_rows_per_group = [x.shape[0] for x in data_list]
        expected_length = self.d_grid.shape[0]

        if not all(n == expected_length for n in n_rows_per_group):
            raise ValueError("All groups/replicates must have the same vector length as the domain")

        self._function_subsetter()

        self.n_domain_points = len(range(self.lb_index, self.ub_index + 1))

        if self.n_domain_points < 1000:
            warnings.warn(f'Functional data has a resolution of {self.n_domain_points} elements. It is recommended to have a resolution of at least 1000 elements for the convergence of the F-ANOVA p-values')

        # Subset  data
        for k in range(self._groups.k):
            self.data.append(data_list[k][self.lb_index : self.ub_index + 1, :]) # exclusive at the upper bound

        self._validate_subgroup_indicator(subgroup_indicator)
        if self._groups.subgroup_indicator is None: # One Way ANOVA set up
            if  self._labels.group:
                assert len(self._labels.group) == self._groups.k, "Each Group Must Have Exactly One Label Associated to it"
                self._labels.generic_group = False
            else:
                self._labels.group = [f"{i+1}" for i in range(self._groups.k)] #Automatically Assign Group Labels
                self._labels.generic_group = True
        else:  # Two Way ANOVA set up
            self._setup_twoway()  # Creates Indicator Matrices and default Labels
            self._n_ii_generator()  # Creates Secondary Size Array


    def plot_means(self,
                   plot_type: str = 'FAMILY',
                   subgroup_indicator: Optional[Union[np.ndarray, List[np.ndarray]]] = None,
                   group_labels: Optional[List[str]] = None,
                   primary_labels: Optional[List[str]] = None,
                   secondary_labels: Optional[List[str]] = None,
                   domain_units: None|str = None,
                   domain_label: None|str = None,
                   response_units: None|str = None,
                   response_label: None|str = None,
                   observation_size_label: bool = True,
                   x_scale: str = 'linear',
                   y_scale: str = 'linear',
                   data_transparency: float = 0.1,
                   legend_transparency: float = 0.3333,
                   data_line_width: float = 1.75,
                   mean_line_width: float = 5,
                   font_size: int = 18,
                   title_labels: Optional[Any] = None,
                   save_path: str = '',
                   legend_location: str = 'best',
                   num_columns: int = 1,
                   legend_title: str = '',
                   new_colors: Optional[np.ndarray] = None,
                   position: Tuple[int, int, int, int] = (90, 90, 1400, 800)) -> Tuple[Any, Any]:

        if subgroup_indicator is not None:
            self._validate_subgroup_indicator(subgroup_indicator, assign=True) # Verify First
            self._setup_twoway()  # Creates Indicator Matrices and default Labels
            self._n_ii_generator()  # Creates Secondary Size Array

        self._verifyLabels(group_labels, primary_labels, secondary_labels) # Requires subgroup_indicator check first
        self._validate_domain_response_labels(domain_units, domain_label, response_units, response_label)


        # Build updates dictionary from user args
        updates = dict(
            x_scale=x_scale,
            y_scale=y_scale,
            data_transparency=data_transparency,
            legend_transparency=legend_transparency,
            data_line_width=data_line_width,
            mean_line_width=mean_line_width,
            font_size=font_size,
            title_labels=title_labels,
            save_path=save_path,
            legend_location=legend_location,
            num_columns=num_columns,
            legend_title=legend_title,
            new_colors=new_colors,
            position=position,
        )

        # Update only changed values
        self.plottingOptions.update_from_dict(updates)

        # Normalize plot_type once
        plot_type = plot_type.upper()

        if self._groups.subgroup_indicator is None:
            # One-way analysis
            if plot_type != 'FAMILY':
                raise ValueError(
                    f"Invalid plot_type '{plot_type}'. "
                    "For one-way analyses, only plot_type='FAMILY' is allowed."
                )
        else:
            # Two-way analysis
            valid_plot_types = ['PRIMARY', 'SECONDARY', 'INTERACTION', 'FAMILY']
            if plot_type not in valid_plot_types:
                raise ValueError(
                    f"Invalid plot_type '{plot_type}'. "
                    f"Valid options for two-way analyses are {valid_plot_types}."
                )

        return plotting.plot_means(self, plot_type)

    def plot_covariances(self,
                        plot_type:str='FAMILY',
                        subgroup_indicator: Optional[Union[np.ndarray, List[np.ndarray]]] = None,
                        group_labels: Optional[List[str]] = None,
                        primary_labels: Optional[List[str]] = None,
                        secondary_labels: Optional[List[str]] = None,
                        domain_units: None|str = None,
                        domain_label: None|str = None,
                        response_units: None|str = None,
                        response_label: None|str = None,
                        x_scale: str = 'linear',
                        y_scale: str = 'linear',
                        color_scale: str = 'linear',
                        font_size: int = 18,
                        title_labels: Optional[Any] = None,
                        save_path: str = '',
                        position: Tuple[int, int, int, int] = (90, 257, 2000, 800)) -> Any:

        if subgroup_indicator is not None:
            self._validate_subgroup_indicator(subgroup_indicator, assign=True) # Verify First
            self._setup_twoway()  # Creates Indicator Matrices and default Labels
            self._n_ii_generator()  # Creates Secondary Size Array

        self._verifyLabels(group_labels, primary_labels, secondary_labels) # Requires subgroup_indicator check first

        # return plotting.plot_covariances(self, plot_type, subgroup_indicator, group_labels, primary_labels, secondary_labels, x_scale, y_scale, color_scale, domain_units_label, response_units_label, title_labels, save_path, position)
        self._validate_domain_response_labels(domain_units, domain_label, response_units, response_label)

        updates = dict(
            x_scale=x_scale,
            y_scale=y_scale,
            color_scale=color_scale,
            font_size=font_size,
            title_labels=title_labels,
            save_path=save_path,
            position=position
        )

        # update only changed values
        self.plottingOptions.update_from_dict(updates)

        return plotting.plot_covariances(self, plot_type)

    def _plot_test_stats(self, p_value, null_dist, test_stat, test_name,
                         scedasticity:str, k:int, N:int|None=None):
        return plotting.plot_test_stats(self, p_value, null_dist, test_stat, test_name, scedasticity, k, N)

    def _run_oneway(self, *args, **kwargs):
        return oneway.run_oneway(self, *args, **kwargs)

    def _run_onewayBF(self, *args, **kwargs):
        return oneway.run_onewayBF(self, *args, **kwargs)

    def _run_twoway(self, *args, **kwargs):
        return twoway.run_twoway(self, *args, **kwargs)

    def _run_twowayBF(self, *args, **kwargs):
        return twoway.run_twowayBF(self, *args, **kwargs)

    def oneway(self,
               n_boot: Optional[int] = None,
               n_simul: Optional[int] = None,
               alpha: Optional[float] = None,
               seed: Optional[Union[int, np.random.Generator]] = None,
               methods: Optional[Sequence[str]] = None,
               hypothesis: Optional[Sequence[str]] = None,
               show_simul_plots: Optional[bool] = None,
               group_labels: Optional[Union[List[str], Tuple[str, ...]]] = None):
        
        self._validate_stat_inputs(alpha, n_boot, n_simul, seed, methods, hypothesis, design='oneway')
        self._validate_plotting_inputs(show_simul_plots, group_labels)

        n_methods = len(self._methods.anova_methods_used)

        eta_i, eta_grand, build_Covar_star = utils.compute_group_means(self._groups.k, self.n_domain_points, self.data, self.n_i, self.N)
        H0 = self._computeSSH_and_pairs(eta_i, eta_grand)

        if np.all(H0.SSH_t < np.finfo(float).eps):
            warnings.warn("Pointwise between-subject variation is 0. Check for duplicated data matrices.")


        # Compute estimated covariance
        gamma_hat = (1 / (self.N - self._groups.k)) * (build_Covar_star @ build_Covar_star.T)  #changed transpose order

        # Only Positive eigen values
        eig_gamma_hat = np.linalg.eigvalsh(gamma_hat)
        eig_gamma_hat = eig_gamma_hat[eig_gamma_hat > 0]


        # Compute SSE(t)
        SSE_t = (self.N - self._groups.k) * np.diag(gamma_hat) # Pointwise within-subject  (Group/Categorical) variations

        match self.hypothesis:
            case 'FAMILY':
                 self._tables.oneway = pd.DataFrame({
                     'Family-Wise Method': list(self._methods.anova_methods_used),
                     'Test-Statistic': [np.nan] * n_methods,
                     'P-Value': [np.nan] * n_methods,
                     'Verdict': [''] * n_methods,
                     'Parameter 1 Name': [''] * n_methods,
                     'Parameter 1 Value': [None] * n_methods,
                     'Parameter 2 Name': [''] * n_methods,
                     'Parameter 2 Value': [None] * n_methods})
                 self._labels.hypothesis = ['Family']
            case 'PAIRWISE':
                self._tables.oneway = pd.DataFrame({'Hypothesis': H0.pair_vec})
            case _:
                pass

        T_hypothesis = pd.DataFrame({'Hypothesis': H0.pair_vec})
        params = self._setup_oneway(H0, SSE_t, gamma_hat)

        p_value_matrix = self._run_oneway(eig_gamma_hat, eta_i, params, H0)

        self._prep_tables('oneway', p_value_matrix, T_hypothesis=T_hypothesis, test_stat=None)

        if self.verbose:
            self._data_summary_report_one_way(ANOVA_TYPE='homoskedastic')
            self._show_table(self._tables.oneway)

    def oneway_bf(self,
               n_boot: Optional[int] = None,
               n_simul: Optional[int] = None,
               alpha: Optional[float] = None,
               seed: Optional[Union[int, np.random.Generator]] = None,
               methods: Optional[Sequence[str]] = None,
               hypothesis: Optional[Sequence[str]] = None,
               show_simul_plots: Optional[bool] = None,
               group_labels: Optional[Union[List[str], Tuple[str, ...]]] = None):
        
        # Use instance defaults unless overridden
        if n_boot is not None:
            self.n_boot = int(n_boot)

        if n_simul is not None:
            self.n_simul = int(n_simul)

        if alpha is not None:
            self.alpha = alpha
            
        self._validate_stat_inputs(alpha, n_boot, n_simul, seed, methods, hypothesis, design='oneway')
        self._validate_plotting_inputs(show_simul_plots, group_labels)

        n_methods = len(self._methods.anova_methods_used)

        yy = np.vstack([arr.T for arr in self.data])

        pair_vec = []

        # Set up Hypothesis
        match self.hypothesis:
            case "PAIRWISE":
                C = utils.construct_pairwise_contrast_matrix(self._groups.k)
                n_tests = C.shape[0]
                if self._labels.group is None:
                    raise ValueError("Group labels must be provided.")
                if len(self._labels.group) != self._groups.k:
                    raise ValueError(
                        f"Each group must have exactly one label. "
                        f"Got {len(self._labels.group)} labels for {self._groups.k} groups.")
                for cc in range(n_tests):
                    # Find the indices where C[cc, :] is True
                    idx = np.where(C[cc])[0]
                    t1, t2 = self._labels.group[idx[0]], self._labels.group[idx[1]]
                    pair_label = f"{t1} & {t2}"
                    pair_vec.append(pair_label)

            case "FAMILY":
                n_tests = 1
                pair_vec.append("FAMILY")
                self._labels.hypothesis = pair_vec
                k = self._groups.k
                C = np.hstack([np.eye(k - 1), -np.ones((k - 1, 1))])

                self._tables.oneway_bf = pd.DataFrame({
                                    'Family-Wise Method': self._methods.anova_methods_used,
                                    'Test-Statistic': [np.nan] * n_methods,
                                    'P-Value': [np.nan] * n_methods,
                                    'Verdict': [None] * n_methods  # Empty strings for verdict (like MATLAB strings)
                                })
            case _:
                raise ValueError("Unknown Hypothesis provided")

        # Iterate over Methods
        # Convert pair_vec to a DataFrame
        T_hypothesis = pd.DataFrame({'Hypothesis': pair_vec})

        # Create NaN matrices
        p_value_matrix = np.full((n_tests, n_methods), np.nan)
        test_stat = np.full((1, n_methods), np.nan)

        counter = 0
        c = 0  # Assuming Equality to each other and not a generic constant
        statistic = np.nan

        for method in self._methods.anova_methods_used:
            p_value = np.zeros(n_tests)

            for ii in range(n_tests):
                if self.hypothesis == 'FAMILY':
                    C_input = C
                elif self.hypothesis == 'PAIRWISE':
                    C_input = C[ii, :]
                    self._labels.hypothesis = pair_vec[ii]
                else:
                    raise ValueError(f"Unsupported hypothesis type: {self.hypothesis}")

                p_value[ii], statistic = self._run_onewayBF(method, yy, C_input, c, self.rng)

            p_value_matrix[:, counter] = p_value
            test_stat[0, counter] = statistic
            counter += 1

        self._prep_tables('oneway_bf', p_value_matrix, T_hypothesis, test_stat)

        if self.verbose:
            self._data_summary_report_one_way(ANOVA_TYPE='heteroskedastic')
            self._show_table(self._tables.oneway_bf)

    def twoway(
        self,
        n_boot: Optional[int] = None,
        n_simul: Optional[int] = None,
        alpha: Optional[float] = None,
        seed: Optional[Union[int, np.random.Generator]] = None,
        methods: Optional[Sequence[str]] = None,
        hypothesis: Optional[Sequence[str]] = None,
        subgroup_indicator: Union[np.ndarray, List[np.ndarray], None] = None,
        contrast: Optional[Union[np.ndarray, List[int], None]] = None,
        primary_labels: Optional[Sequence[str]] = None,
        secondary_labels: Optional[Sequence[str]] = None,
        weights: Optional[Literal["proportional", "uniform"]] = None
    ):
        self._validate_stat_inputs(alpha, n_boot, n_simul, seed, methods, hypothesis,  design='twoway')
        self._validate_twoway_inputs(contrast, weights)
        self._validate_plotting_inputs(primary_labels, secondary_labels)

        if self._groups.subgroup_indicator is None:
            raise ValueError('subgroup_indicator must be provided for twoway ANOVAs')
        else:
            self._setup_twoway()  # already validates subgroup_indicator
            self._setup_twoway_h0()

        yy = np.vstack([arr.T for arr in self.data])

        C, n_tests = self._compute_contrast_matrix()
        C = np.asarray(C)
        pair_vec = self._setup_twoway_labels_and_tables(C, n_tests, scedasticity='homoscedastic')

        T_hypothesis = pd.DataFrame({'Hypothesis': pair_vec})
        n_methods = len(self._methods.anova_methods_used)
        # Always initialize analysis state
        p_value_matrix = np.full((n_tests, n_methods), np.nan)
        test_stat = np.full((1, n_methods), np.nan)
        counter = 0
        statistic = 0
        c = 0

        for counter, method in enumerate(self._methods.anova_methods_used):
            p_values = np.zeros(n_tests)

            for ii in range(n_tests):
                match self.hypothesis:
                    case "PAIRWISE":
                        C_input = C[ii, :]
                        self._labels.hypothesis = pair_vec[ii]

                    case "INTERACTION" | "PRIMARY" | "SECONDARY" | "FAMILY":
                        C_input = C

                    case _:
                        C_input = C  # for CUSTOM or any other hypothesis

                # Run the TwoWay test
                p_value, statistic = self._run_twoway(method, yy, C_input, self.rng)
                p_values[ii] = p_value

            # Store results
            p_value_matrix[:, counter] = p_values
            test_stat[0, counter] = statistic

        self._prep_tables('twoway', p_value_matrix, T_hypothesis, test_stat)
        if self.verbose:
            self._data_summary_report_two_way(ANOVA_TYPE='homoskedastic')
            self._show_table(self._tables.twoway)

    def twoway_bf(
        self,
        n_boot: Optional[int] = None,
        n_simul: Optional[int] = None,
        alpha: Optional[float] = None,
        seed: Optional[Union[int, np.random.Generator]] = None,
        methods: Optional[Sequence[str]] = None,
        hypothesis: Optional[Sequence[str]] = None,
        subgroup_indicator: Union[np.ndarray, List[np.ndarray], None] = None,
        contrast: Optional[np.ndarray] = None,
        primary_labels: Optional[Sequence[str]] = None,
        secondary_labels: Optional[Sequence[str]] = None,
        weights: Optional[Literal["proportional", "uniform"]] = None
    ):
        self._validate_stat_inputs(alpha, n_boot, n_simul, seed, methods, hypothesis, design='oneway')
        self._validate_twoway_inputs(contrast, weights)
        self._validate_plotting_inputs(primary_labels, secondary_labels)

        if self._groups.subgroup_indicator is None:
            raise ValueError('subgroup_indicator must be provided for twoway ANOVAs')
        else:
            self._setup_twoway()  # already validates subgroup_indicator
            self._setup_twoway_h0()

        yy = np.vstack([arr.T for arr in self.data])

        C, n_tests = self._compute_contrast_matrix()
        C = np.asarray(C)
        pair_vec = self._setup_twoway_labels_and_tables(C, n_tests, scedasticity='heteroscedastic')

        T_hypothesis = pd.DataFrame({'Hypothesis': pair_vec})
        n_methods = len(self._methods.anova_methods_used)
        # Always initialize analysis state
        p_value_matrix = np.full((n_tests, n_methods), np.nan)
        test_stat = np.full((1, n_methods), np.nan)
        counter = 0
        statistic = 0
        c = 0

        for counter, method in enumerate(self._methods.anova_methods_used):
            p_values = np.zeros(n_tests)

            for ii in range(n_tests):
                match self.hypothesis:
                    case "PAIRWISE":
                        C_input = C[ii, :]
                        self._labels.hypothesis = pair_vec[ii]

                    case "INTERACTION" | "PRIMARY" | "SECONDARY" | "FAMILY":
                        C_input = self.sanitize_contrast(C)

                    case _:
                        C_input = self.sanitize_contrast(C)  # for CUSTOM or any other hypothesis

                # Run the TwoWay test
                p_value, statistic = self._run_twowayBF(method, yy, C_input, c)
                p_values[ii] = p_value

            # Store results
            p_value_matrix[:, counter] = p_values
            test_stat[0, counter] = statistic

        self._prep_tables('twoway_bf', p_value_matrix, T_hypothesis, test_stat)
        if self.verbose:
            self._data_summary_report_two_way(ANOVA_TYPE='heteroscedastic')
            self._show_table(self._tables.twoway_bf)

    def _prep_tables(self, anova_method, p_value_matrix, T_hypothesis, test_stat):
        match anova_method:
            case 'oneway':
                match self.hypothesis:
                    case "PAIRWISE":
                        T_p_value = pd.DataFrame(p_value_matrix, columns=self._methods.anova_methods_used)
                        self._tables.oneway = pd.concat([T_hypothesis, T_p_value], axis=1)

                    case "FAMILY":
                        assert self._tables.oneway is not None, 'Oneway table is not properly setup'

                        self._tables.oneway["P-Value"] = np.array(p_value_matrix).flatten()

                        signif_results = self._tables.oneway["P-Value"] < self.alpha

                        self._tables.oneway.loc[signif_results, "Verdict"] = "Reject Null Hypothesis for Alternative Hypothesis"
                        self._tables.oneway.loc[~signif_results, "Verdict"] = "Fail to Reject Null Hypothesis"

                        # if self.consistantTable:
                        #     self.OneWay_P_Table = self.OneWay_P_Table.iloc[:, :4]
            case 'oneway_bf':
                match self.hypothesis:
                    case "PAIRWISE":
                        # Use all methods
                        T_p_value = pd.DataFrame(
                            p_value_matrix,
                            columns=self._methods.anova_methods_used
                        )
                        self._tables.oneway_bf = pd.concat([T_hypothesis, T_p_value], axis=1)

                    case "FAMILY":
                        # Keep the full table
                        if not isinstance(self._tables.oneway_bf, pd.DataFrame):
                            raise ValueError('oneway_bf should be a pandas dataframe')

                        self._tables.oneway_bf = self._tables.oneway_bf.copy()

                        self._tables.oneway_bf["P-Value"] = p_value_matrix.flatten()
                        self._tables.oneway_bf["Test-Statistic"] = test_stat.flatten()

                        signif_results = self._tables.oneway_bf["P-Value"] < self.alpha
                        self._tables.oneway_bf.loc[signif_results, "Verdict"] = ("Reject Null Hypothesis for Alternative Hypothesis")
                        self._tables.oneway_bf.loc[~signif_results, "Verdict"] = ("Fail to Reject Null Hypothesis")
            case 'twoway':
                    match self.hypothesis:
                        case "PAIRWISE":
                            # Create DataFrame of p-values
                            T_p_value = pd.DataFrame(p_value_matrix, columns=self._methods.anova_methods_used)

                            # Combine into full results table
                            self._tables.twoway = pd.concat([T_hypothesis, T_p_value], axis=1)

                        case "FAMILY" | "PRIMARY" | "SECONDARY" | "INTERACTION" | "CUSTOM":
                            if not isinstance(self._tables.twoway, pd.DataFrame):
                                raise ValueError('two_way should be a pandas dataframe')

                            # Update columns
                            self._tables.twoway["P-Value"] = p_value_matrix.T.flatten()
                            self._tables.twoway["Test-Statistic"] = test_stat.T.flatten()

                            # Add verdicts based on alpha
                            significant = self._tables.twoway["P-Value"] < self.alpha
                            self._tables.twoway.loc[significant, "Verdict"] = "Reject Null Hypothesis for Alternative Hypothesis"
                            self._tables.twoway.loc[~significant, "Verdict"] = "Fail to Reject Null Hypothesis"
                        case _:
                            raise ValueError(f"Unsupported Hypothesis: {self.hypothesis}")

            case 'twoway_bf':
                     match self.hypothesis:
                        case "PAIRWISE" :
                            # Create DataFrame of p-values
                            T_p_value = pd.DataFrame(p_value_matrix, columns=self._methods.anova_methods_used)

                            # Combine into full results table
                            self._tables.twoway_bf = pd.concat([T_hypothesis, T_p_value], axis=1)

                        case "FAMILY" | "PRIMARY" | "SECONDARY" | "INTERACTION" | "CUSTOM":
                            if not isinstance(self._tables.twoway_bf, pd.DataFrame):
                                raise ValueError('two_way should be a pandas dataframe')

                            # Update columns
                            self._tables.twoway_bf["P-Value"] = p_value_matrix.T.flatten()
                            self._tables.twoway_bf["Test-Statistic"] = test_stat.T.flatten()

                            # Add verdicts based on alpha
                            significant = self._tables.twoway_bf["P-Value"] < self.alpha
                            self._tables.twoway_bf.loc[significant, "Verdict"] = "Reject Null Hypothesis for Alternative Hypothesis"
                            self._tables.twoway_bf.loc[~significant, "Verdict"] = "Fail to Reject Null Hypothesis"
                        case _:
                            raise ValueError(f"Unsupported Hypothesis: {self.hypothesis}")

    def _show_table(self, table_to_show):
        temp_table = table_to_show.copy()

        if self.hypothesis == "PAIRWISE":
            all_var_names = list(temp_table.columns)
            n_items = len(temp_table)

            # Format all columns except the first
            for col_name in all_var_names[1:]:
                temp_table[col_name] = temp_table[col_name].apply(
                    lambda x: f"{x:6.{self._tables.sig_figs}f}"
                )

            # Reorder columns according to self.ANOVA_Methods
            method_order = {method: i for i, method in enumerate(self._methods.anova_methods)}
            reordered_cols = [all_var_names[0]] + sorted(
                all_var_names[1:], key=lambda col: method_order.get(col, float('inf'))
            )
            temp_table = temp_table[reordered_cols]

        else:  # "FAMILY" | "PRIMARY" | "SECONDARY"
            n_items = len(temp_table)

            # Format 'Test-Statistic' column
            temp_table["Test-Statistic"] = temp_table["Test-Statistic"].apply(
                lambda x: f"{x:6.{self._tables.sig_figs}f}"
            )

            # Format 'P-Value' column
            temp_table["P-Value"] = temp_table["P-Value"].apply(
                lambda x: f"{x:6.{self._tables.sig_figs}f}"
            )


            # Conditionally format Parameter 1/2 Value columns
            for col in ["Parameter 1 Value", "Parameter 2 Value"]:
                if col in temp_table.columns:
                    temp_table[col] = temp_table[col].apply(
                        lambda x: f"{x:6.{self._tables.sig_figs}f}" if isinstance(x, float) else x
                    )

            # Sort rows based on order in self.ANOVA_Methods
            method_order = {method: i for i, method in enumerate(self._methods.anova_methods)}
            temp_table = temp_table.sort_values(
                by=temp_table.columns[0],  # first column, e.g. "Family-Wise Method"
                key=lambda col: col.map(method_order)
            ).reset_index(drop=True)

        print(temp_table)

    def _data_summary_report_one_way(self, ANOVA_TYPE):
        n_groups = len(self.n_i)

        Mystring = f"\nOne-Way {ANOVA_TYPE} F-ANOVA Data Summary:\n\n"
        Mystring += f"Hypothesis Type = {self.hypothesis.upper()}\n"
        Mystring += f"Confidence Level = {(1 - self.alpha) * 100:.3f} %\n"
        Mystring += f"Number of Observations Total = {self.N}\n"
        Mystring += f"Number of Points in Domain = {self.n_domain_points}\n"
        Mystring += f"Number of Groups = {n_groups}\n"
        Mystring += f"Domain Range = [{self.d_grid[0]:.3f}, {self.d_grid[-1]:.3f}]\n"
        Mystring += f"Domain Subset = [{self.grid_bounds[0]:.3f}, {self.grid_bounds[-1]:.3f}]\n"
        Mystring += f"Group Observation Size: [{', '.join(str(x) for x in self.n_i)}]\n"

        if self._labels.group:
            Mystring += f"Group Labels: [{', '.join(self._labels.group)}]\n"

        Mystring += "\n"
        print(Mystring)

    def _data_summary_report_two_way(self, ANOVA_TYPE):
        # Build Secondary Factor Observation Size string (B_obs_string)
        if self.n_ii is None:
            raise ValueError('n_ii attribute is None when it should be populated for this report')

        B_obs_string_parts = []
        for k in range(self._groups.A):
            obs = ' '.join(str(x) for x in self.n_ii[k])
            B_obs_string_parts.append(f"[{obs}]")
        B_obs_string = ', '.join(B_obs_string_parts)

        # Start building the summary string
        Mystring = f"\nTwo-Way {ANOVA_TYPE} F-ANOVA Data Summary:\n\n"
        Mystring += f"Hypothesis Type = {self.hypothesis.upper()}\n"
        Mystring += f"Confidence Level = {(1 - self.alpha) * 100:.3f} %\n"
        Mystring += f"Number of Observations Total = {self.N}\n"
        Mystring += f"Number of Points in Domain = {self.n_domain_points}\n"
        Mystring += f"Number of Groups within Primary Factor = {self._groups.A}\n"
        Mystring += f"Number of Groups within Secondary Factor= {self._groups.B}\n"
        Mystring += f"Number of Total Groups = {self._groups.AB}\n"
        Mystring += f"Domain Range = [{self.d_grid[0]:.3f}, {self.d_grid[-1]:.3f}]\n"
        Mystring += f"Domain Subset = [{self.grid_bounds[0]:.3f}, {self.grid_bounds[-1]:.3f}]\n"
        Mystring += f"Primary Factor Observation Size: [{', '.join(str(x) for x in self.n_i)}]\n"
        Mystring += f"Secondary Factor Observation Size: [{B_obs_string}]\n"

        if self._labels.primary:
            Mystring += f"Primary Factor Labels: [{', '.join(self._labels.primary)}]\n"

        if self._labels.secondary:
            Mystring += f"Secondary Factor Labels: [{', '.join(self._labels.secondary)}]\n"

        Mystring += "\n"
        print(Mystring)

    def _cast_anova_methods(self, method):
        """
        Filters self._methods.anova_methods_used to keep only valid methods.
        Warns about any unrecognized method names.
        """
        valid_methods = set(self._methods.anova_methods)
        original = sorted(list(method), reverse=True) # Maintain L2 first than F-test

        # Keep only valid methods (case-insensitive match)
        filtered = [m for m in original if m in valid_methods]
        excluded = [m for m in original if m not in valid_methods]

        self._methods.anova_methods_used = tuple(filtered)

        if excluded:
            warnings.warn(f"These unknown methods were excluded: {' & '.join(excluded)}")

        assert len(filtered) >= 1, (
            "No ANOVA methods were selected!\n"
            f"Must be at least one of the following: {', '.join(self._methods.anova_methods)}"
        )

    def _validate_instantiation_inputs(self):

        # Validate bounds
        if not isinstance(self.grid_bounds, tuple) or len(self.grid_bounds) != 2:
            raise ValueError(f"F-ANOVA bounds must be a tuple of length 2, but got {type(self.grid_bounds).__name__} with value {self.grid_bounds}")
        if not all(isinstance(x, (int, float)) for x in self.grid_bounds):
            raise ValueError(f"F-ANOVA bounds must contain numeric values, but got {self.grid_bounds}")

        # Validate d_grid
        self.d_grid = self._cast_to_1D(self.d_grid)
        #TODO need to validate more inputs

    def _validate_stat_inputs(
        self,
        alpha,
        n_boot,
        n_simul,
        seed,
        methods,
        hypothesis,
        design: str,   # <-- explicit
    ):
        # Resolve effective values (do NOT mutate yet)
        alpha_eff = self.alpha if alpha is None else alpha
        n_boot_eff = self.n_boot if n_boot is None else int(n_boot)
        n_simul_eff = self.n_simul if n_simul is None else int(n_simul)

        # ---- Validate alpha ----
        if not (0 < alpha_eff < 1):
            raise ValueError(f"alpha must be between 0 and 1, got {alpha_eff}")

        # ---- Validate boot/simul ----
        if not isinstance(n_boot_eff, int) or n_boot_eff <= 0:
            raise ValueError(f"n_boot must be a positive integer, got {n_boot_eff}")

        if not isinstance(n_simul_eff, int) or n_simul_eff <= 0:
            raise ValueError(f"n_simul must be a positive integer, got {n_simul_eff}")

        # ---- Validate hypothesis ----
        if hypothesis is not None:
            if not isinstance(hypothesis, str):
                raise TypeError(f"'hypothesis' must be a string, got {type(hypothesis).__name__}")

            hypothesis = hypothesis.upper().strip()

            if design == "oneway":
                valid = self._labels.H0_OneWay
            elif design == "twoway":
                valid = self._labels.H0_TwoWay
            else:
                raise ValueError(f"Unknown design type: {design}")

            if hypothesis not in valid:
                raise ValueError(f"Invalid hypothesis: {hypothesis}. Must be one of {valid}")

            self.hypothesis = hypothesis

        # ---- Validate methods ----
        if methods is not None:
            upper_cased_methods = tuple(s.upper() for s in self._methods.anova_methods)
            for m in methods:
                if m.upper() not in upper_cased_methods:
                    raise ValueError(
                        f"Invalid method: {m}. Must be one of {self._methods.anova_methods}"
                    )
            self._cast_anova_methods(methods)
        else:
            self._methods.anova_methods_used = self._methods.anova_methods

        # ---- RNG ----
        if seed is not None and not isinstance(seed, (int, np.random.Generator)):
            raise TypeError(f"'seed' must be an integer or a RNG, got {type(seed).__name__}")

        self.rng = self._get_rng(seed)

        # ---- Commit validated values ----
        self.alpha = alpha_eff
        self.n_boot = n_boot_eff
        self.n_simul = n_simul_eff

    def _validate_subgroup_indicator(
        self,
        subgroup_indicator: Optional[Union[np.ndarray, List[np.ndarray]]] = None,
        *,
        assign: bool = False,
    ) -> None:
        if subgroup_indicator is None:
            return

        # ---- single array case: expects overall N ----
        if isinstance(subgroup_indicator, np.ndarray):
            if subgroup_indicator.shape not in ((self.N,), (self.N, 1)):
                raise ValueError(
                    f"subgroup_indicator has shape {subgroup_indicator.shape}; "
                    f"expected {(self.N,)} or {(self.N, 1)} (N={self.N})."
                )

        # ---- list case: expects one per group, matching n_i 1-for-1 ----
        elif isinstance(subgroup_indicator, list):
            if not subgroup_indicator:
                raise ValueError("subgroup_indicator list must not be empty")

            if not all(isinstance(x, np.ndarray) for x in subgroup_indicator):
                raise TypeError("subgroup_indicator must be a np.ndarray or a list of np.ndarray")

            k = self.groups.k
            if len(subgroup_indicator) != k:
                raise ValueError(
                    f"subgroup_indicator list length ({len(subgroup_indicator)}) "
                    f"must match the number of groups (k={k})."
                )

            if len(self.n_i) != k:
                raise ValueError(f"n_i length ({len(self.n_i)}) must match k={k}.")

            bad = [
                i for i, (x, n) in enumerate(zip(subgroup_indicator, self.n_i))
                if x.shape not in ((n,), (n, 1))
            ]
            if bad:
                i = bad[0]
                n = self.n_i[i]
                raise ValueError(
                    f"subgroup_indicator[{i}] has shape {subgroup_indicator[i].shape}; "
                    f"expected {(n,)} or {(n, 1)} to match n_i[{i}]={n}."
                )

        else:
            raise TypeError("subgroup_indicator must be a np.ndarray or a list of np.ndarray")

        # ---- only reached if no errors occurred ----
        if assign:
            self._groups.subgroup_indicator = subgroup_indicator

            if self._labels.group is not None: # transition from One-Way to Two-Way Analyses via subgroup
                self._labels.primary = self._labels.group   # Move group now to primary
                self._labels.group = None # Force to None

    def _validate_plotting_inputs(self, show_simul_plots=None, group_labels=None, primary_labels=None, secondary_labels=None):
        if show_simul_plots is not None:
            if not isinstance(show_simul_plots, bool):
                raise TypeError(f"'show_simul_plots' must be True or False, got {type(show_simul_plots).__name__}")
            self.show_simul_plots = show_simul_plots
        else:
            self.show_simul_plots = False # default

        if group_labels is not None:
            if not isinstance(group_labels, (list, tuple)):
                raise TypeError(f"'group_labels' must be a list or tuple, got {type(group_labels).__name__}")
            if not all(isinstance(label, str) for label in group_labels):
                raise ValueError("All labels in 'group_labels' must be strings")
            if not len(group_labels) == self._groups.k:
                raise ValueError(f"length of 'group_labels' {len(group_labels)} must match number of groups {self._groups.k}")
            self._labels.group = list(group_labels)
            self._labels.generic_group = False

        if primary_labels is not None:
            if not isinstance(primary_labels, (list, tuple)):
                raise TypeError(f"'primary_labels' must be a list or tuple, got {type(primary_labels).__name__}")
            if not all (isinstance(label, str) for label in primary_labels):
                raise ValueError("All labels in 'primary_labels' must be strings")

        self.primary_labels = primary_labels

        if secondary_labels is not None:
            if not isinstance(secondary_labels, (list, tuple)):
                raise TypeError(f"'secondary_labels' must be a list or tuple, got {type(secondary_labels).__name__}")
            if not all (isinstance(label, str) for label in secondary_labels):
                raise ValueError("All labels in 'secondary_labels' must be strings")

        self.secondary_labels = secondary_labels

    def _validate_twoway_inputs(self, contrast, weights):
        # Check contrast variable
        if contrast is not None:
            if isinstance(contrast, list):
                all_are_int = all(isinstance(x, int) for x in contrast)
                if not all_are_int:
                    raise TypeError(f"All element of the list must be python integer types")
            elif isinstance(contrast, np.ndarray):
                all_are_int = np.issubdtype(contrast.dtype, np.integer)
                if not all_are_int:
                    raise TypeError(f"All element of the NumPy array must be integer types")
                if contrast.ndim not in [1,2]:
                    raise ValueError(f"'contrast' must be a 1D or 2D array, got {contrast.ndim}")
            else:
                raise TypeError(f"'contrast' must be a NumPy array or list of integers, got {type(contrast).__name__}")

            # if self._groups.A == self._groups.B:
            #     if len(contrast) != self._groups.A:
            #         raise ValueError(f'Size of Contrast Vector={len(contrast)} doesnt match the number of levels in either the Primary or Secondary Factors, n={self._groups.A}.')
            
            #     indx, tf = self.select_factor()
                
            #     if tf:
            #         match indx:
            #             case 1:
            #                 print('Contrast Vector Applied to Primary Factor')
            #                 self._groups.contrast_factor= 1
            #             case 2:
            #                 print('Contrast Vector Applied to Secondary Factor')
            #                 self._groups.contrast_factor = 2
            #     else:
            #         raise ValueError('No Selection Made')
                        
            
            # else:  # Different sizes Check size of contrast
            #     if len(contrast) == self._groups.A:
            #         print('Contrast Vector Applied to Primary Factor')
            #         self._groups.contrast_factor = 1
            #     elif len(contrast) == self._groups.B:
            #         print('Contrast Vector Applied to Secondary Factor')
            #         self._groups.contrast_factor = 2
            #     else:
            #         raise ValueError(f'Size of Contrast Vector={len(contrast)} doesnt match the number of levels in either the Primary={self._groups.A} or Secondary={self._groups.B} Factors.')
            self._groups.contrast = contrast
        elif contrast is not None:
            warnings.warn('Ignoring Contrast Vector. Contrast Vector only used when hypothesis="CUSTOM"')

        

        if weights is not None:
            if not isinstance(weights, str):
                raise TypeError(f"'weights' must be a string, got {type(weights).__name__}")
            # weights type is currently checked twice, once in _validate_twoway_inputs() and once in run_twoway()
            if weights.upper() not in ["UNIFORM", "PROPORTIONAL"]:
                raise ValueError(f"Unsupported weight type: {weights}, must either be 'UNIFORM' or 'PROPORTIONAL'")

            self.weights = weights.upper()
        else:
            self.weights = "PROPORTIONAL"
            
    @staticmethod
    def select_factor():
        options = ["Primary", "Secondary"]

        while True:
            print("\nSelect a Factor:")
            for i, option in enumerate(options, 1):
                print(f"{i}. {option}")

            choice = input("Enter number (or press Enter to cancel): ").strip()

            if choice == "":
                return None, False

            if choice.isdigit() and 1 <= int(choice) <= len(options):
                return options[int(choice) - 1], True

            print("Invalid selection. Try again.")

    def _validate_domain_response_labels(self, domain_units: None|str, domain_label: None|str , response_units: None|str , response_label: None|str ):
        for name, value in [
        ("domain_units", domain_units),
        ("domain_label", domain_label),
        ("response_units", response_units),
        ("response_label", response_label),]:

            if value is not None and not isinstance(value, str):
                raise TypeError(f"{name} must be a string or None, got {type(value).__name__}.")
            if  isinstance(value, str):
                match name:
                    case 'domain_units':
                        self._units.domain = value
                    case 'domain_label':
                        self._labels.domain = value
                    case 'response_units':
                        self._units.response = value
                    case 'response_label':
                        self._labels.response = value

    def _function_subsetter(self):

        lb = np.min(self.grid_bounds)
        ub = np.max(self.grid_bounds)


        if lb > np.max(self.d_grid):
            raise ValueError(f'The lower bound (lb={lb:g}) is greater than the maximum value in self.d_grid ({np.max(self.d_grid):g}).')


        if ub < np.min(self.d_grid):
            raise ValueError(f'The upper bound (ub={ub:g}) is less than the minimum value in self.d_grid ({np.min(self.d_grid):g}).')

        subset_idx = np.where((self.d_grid >= lb) & (self.d_grid <= ub))[0]

        subset_d_grid = self.d_grid[subset_idx]

        local_min_idx = np.argmin(subset_d_grid)
        local_max_idx = np.argmax(subset_d_grid)

        global_min_idx = subset_idx[local_min_idx]
        global_max_idx = subset_idx[local_max_idx]

        self.lb_index = global_min_idx
        self.ub_index = global_max_idx

    def _verifyIndicator(self) -> None:
        si = self._groups.subgroup_indicator
        if si is None:
            raise TypeError("subgroup_indicator is None; expected it to be assigned before _verifyIndicator")

        if isinstance(si, list):
            flat = np.concatenate(si, axis=0)
        elif isinstance(si, np.ndarray):
            flat = si
        else:
            raise TypeError("subgroup_indicator must be a list of np.ndarray or a np.ndarray")

        flat = self._cast_to_1D(flat)

        if flat.shape != (self.N,):
            raise ValueError(f"subgroup_indicator has shape {flat.shape}, expected {(self.N,)}")

        self._groups.subgroup_indicator = flat

    @staticmethod
    def _check_label(labels: Optional[List[str]], name: str, expected: int) -> None:
        if labels is None:
            return

        if not isinstance(labels, list) or not all(isinstance(x, str) for x in labels):
            raise TypeError(f"{name} must be a list[str] or None")

        if len(labels) != expected:
            raise ValueError(f"{name} length ({len(labels)}) must match {expected}.")


    def _verifyLabels(
        self,
        group_labels: Optional[List[str]] = None,
        primary_labels: Optional[List[str]] = None,
        secondary_labels: Optional[List[str]] = None,
    ) -> None:

        if self._groups.subgroup_indicator is None:  # OneWay
            self._check_label(group_labels, "group_labels", self._groups.k)
            if group_labels is not None:
                self._labels.group = group_labels
                self._labels.generic_group = False

        else:  # TwoWay
            self._check_label(primary_labels, "primary_labels", self._groups.A)
            if primary_labels is not None:
                self._labels.primary = primary_labels
                self._labels.generic_group = False

            self._check_label(secondary_labels, "secondary_labels", self._groups.B)
            if secondary_labels is not None:
                self._labels.secondary = secondary_labels
                self._labels.generic_group = False

    def _setup_oneway(self, H0, SSE_t, gamma_hat):
        import numpy as np

        is_family = self.hypothesis == "FAMILY"

        # Initialize
        T_n = np.sum(H0.SSH_t, axis=0)

        if T_n.ndim == 0:
            T_n = T_n.reshape(1, 1)

        F_n = np.zeros(1)
        beta_hat = 0
        kappa_hat = 0
        beta_hat_unbias = 0
        kappa_hat_unbias = 0

        if self._tables.oneway is None:
            raise ValueError('One Way Table wasnt set up properly')

        # FAMILY: Fill Test-Statistic for L2
        if is_family and any("L2" in m for m in self._methods.anova_methods_used):
            mask = [("L2" in m) for m in self._methods.anova_methods_used]
            self._tables.oneway.loc[mask, "Test-Statistic"] = T_n

        # F-test methods
        if any("F" in m for m in self._methods.anova_methods_used):
            numerator = T_n / H0.q
            denominator = np.sum(SSE_t) / (self.N - self._groups.k)
            F_n = numerator / denominator

            if is_family:
                mask = [("F" in m) for m in self._methods.anova_methods_used]
                self._tables.oneway.loc[mask, "Test-Statistic"] = F_n

        # Naive Approx
        if any("Naive" in m for m in self._methods.anova_methods_used):
            beta_hat = utils.beta_hat(gamma_hat)
            kappa_hat = utils.kappa_hat(gamma_hat)

            if is_family:
                mask_1 = np.array(["Naive" in m for m in self._methods.anova_methods_used])
                mask_2 = np.array(["F" in m for m in self._methods.anova_methods_used])

                only_naive = mask_1 & ~mask_2
                self._tables.oneway.loc[only_naive, "Parameter 1 Name"] = "beta"
                self._tables.oneway.loc[only_naive, "Parameter 2 Name"] = "d"
                self._tables.oneway.loc[only_naive, "Parameter 1 Value"] = beta_hat
                self._tables.oneway.loc[only_naive, "Parameter 2 Value"] = H0.q * kappa_hat

                naive_f = mask_1 & mask_2
                d1 = H0.q * kappa_hat
                d2 = (self.N - self._groups.k) * kappa_hat
                self._tables.oneway.loc[naive_f, "Parameter 1 Name"] = "d1"
                self._tables.oneway.loc[naive_f, "Parameter 2 Name"] = "d2"
                self._tables.oneway.loc[naive_f, "Parameter 1 Value"] = d1
                self._tables.oneway.loc[naive_f, "Parameter 2 Value"] = d2

        # BiasReduced Approx
        if any("BiasReduced" in m for m in self._methods.anova_methods_used):
            beta_hat_unbias = utils.beta_hat_unbias(self.N, self._groups.k, gamma_hat)
            kappa_hat_unbias = utils.kappa_hat_unbias(self.N, self._groups.k, gamma_hat)

            if is_family:
                mask_1 = np.array(["BiasReduced" in m for m in self._methods.anova_methods_used])
                mask_2 = np.array(["F" in m for m in self._methods.anova_methods_used])

                only_bias = mask_1 & ~mask_2
                self._tables.oneway.loc[only_bias, "Parameter 1 Name"] = "beta"
                self._tables.oneway.loc[only_bias, "Parameter 2 Name"] = "d"
                self._tables.oneway.loc[only_bias, "Parameter 1 Value"] = beta_hat_unbias
                self._tables.oneway.loc[only_bias, "Parameter 2 Value"] = H0.q * kappa_hat_unbias

                bias_f = mask_1 & mask_2
                d1 = H0.q * kappa_hat_unbias
                d2 = (self.N - self._groups.k) * kappa_hat_unbias
                self._tables.oneway.loc[bias_f, "Parameter 1 Name"] = "d1"
                self._tables.oneway.loc[bias_f, "Parameter 2 Name"] = "d2"
                self._tables.oneway.loc[bias_f, "Parameter 1 Value"] = d1
                self._tables.oneway.loc[bias_f, "Parameter 2 Value"] = d2

        # Return as dictionary
        return AnovaStatistics(T_n, F_n, beta_hat, kappa_hat, beta_hat_unbias,kappa_hat_unbias)

    def _setup_twoway(self):
        self._verifyIndicator()

        subgroup = cast(np.ndarray, self._groups.subgroup_indicator)  #  it's only for static type checking.
        n_unique_indicators = len(np.unique(subgroup))

        self._groups.B = n_unique_indicators


        self._groups.A = self._groups.k
        self._groups.AB = self._groups.A * self._groups.B

        if not self._labels.secondary and not self._labels.primary:
            self.generic_group_labels = True

        if not self._labels.primary:
            self._labels.primary = [f"{i+1}" for i in range(self._groups.k)] # Automatically Assign Primary Numeric Labels (1,2,3,...)

        if not self._labels.secondary:
             self._labels.secondary = [chr(65 + i) for i in range(n_unique_indicators)] # Automatically Assign Alphabetical Labels (A,B,C,...)

        assert len(self._labels.primary) == self._groups.A, "Labels for each Primary factor level must have a one-to-one correspondence to each level"

        assert len(self._labels.secondary) == self._groups.B, "Labels for each Secondary factor level must have a one-to-one correspondence to each level"

        if self._labels.group:
            raise ValueError('TwoWay ANOVA requires using "primary_labels" and "secondary_labels" as input arguments.\nIt doesnt support the "group_labels" argument due to ambiguity.')

    def _setup_twoway_h0(self):
        """
        Sets up Two-Way ANOVA for custom hypothesis testing using a contrast vector.
        """

        if self.hypothesis == "CUSTOM":
            assert self._groups.contrast is not None and len(self._groups.contrast) > 0, \
                "A user-specified Hypothesis requires a non-empty Contrast vector."
            assert np.isclose(np.sum(self._groups.contrast), 0), \
                "Contrast vector must sum to zero."

            contrast_len = len(self._groups.contrast)

            if self._groups.A == self._groups.B:
                assert contrast_len == self._groups.A, \
                    f"Size of Contrast Vector={contrast_len} doesn't match number of levels: n={self._groups.A}"

                # Prompt user for which factor to apply contrast to
                import tkinter as tk
                from tkinter import messagebox, simpledialog

                root = tk.Tk()
                root.withdraw()  # Hide the main window

                factor_choice = simpledialog.askstring(
                    "Factor Selector", "Apply contrast to which factor? [Primary/Secondary]"
                )

                if factor_choice is None:
                    raise SystemExit("No selection made.")

                factor_choice = factor_choice.strip().lower()
                if factor_choice == "primary":
                    print("Contrast Vector applied to Primary Factor")
                    self._groups.contrast_factor = 1

                elif factor_choice == "secondary":
                    print("Contrast Vector applied to Secondary Factor")
                    self._groups.contrast_factor = 2

                else:
                    raise ValueError("Invalid selection. Choose 'Primary' or 'Secondary'.")

            else:
                if contrast_len == self._groups.A:
                    print("Contrast Vector applied to Primary Factor")
                    self._groups.contrast_factor = 1

                elif contrast_len == self._groups.B:
                    print("Contrast Vector applied to Secondary Factor")
                    self._groups.contrast_factor = 2

                else:
                    raise ValueError(
                        f"Contrast Vector length={contrast_len} does not match the number of levels in either "
                        f"Primary={self._groups.A} or Secondary={self._groups.B} Factors."
                    )
        elif self._groups.contrast is not None:
            warnings.warn("Ignoring Contrast Vector. Contrast Vector only used when Hypothesis='CUSTOM'.")

    def _setup_twoway_labels_and_tables(self, C: np.ndarray, n_tests: int, scedasticity: str):
        """
        Sets up hypothesis labels and result tables for Two-Way ANOVA.

        Parameters:
        - C: contrast matrix (for PAIRWISE or CUSTOM)
        - n_tests: number of hypothesis tests
        - scedasticity: 'homoscedastic' or 'heteroscedastic'
        """
        pair_vec = []
        scedasticity = scedasticity.lower().strip()
        use_bf_table = scedasticity == "heteroscedastic"
        n_methods = len(self._methods.anova_methods_used)
        method_col = self._methods.anova_methods_used

        if self.hypothesis == "FAMILY":
            self.hypothesis_LABEL = ""

            table = pd.DataFrame({
                "Family-Wise Method": method_col,
                "Test-Statistic": [np.nan] * n_methods,
                "P-Value": [np.nan] * n_methods,
                "Verdict": [""] * n_methods
            })

            if use_bf_table:
                self._tables.twoway_bf = table
            else:
                self._tables.twoway = table

        elif self.hypothesis == "PAIRWISE":
            all_labels = utils.generate_two_way_comb(self)
            labels = np.asarray(all_labels)
            for cc in range(n_tests):
                indices = C[cc, :].astype(bool)
                picked = labels[indices]          # boolean mask works now
                assert picked.size == 2, f"Expected 2 labels, got {picked.size}: {picked}"
                t1, t2 = picked
                pair_vec.append(f"{t1} & {t2}")

            self.hypothesis_LABEL = np.array(pair_vec)

        elif self.hypothesis in {"INTERACTION", "PRIMARY", "SECONDARY"}:
            label = self.hypothesis.capitalize()
            self.hypothesis_LABEL = label

            table = pd.DataFrame({
                f"{label} Effect": method_col,
                "Test-Statistic": [np.nan] * n_methods,
                "P-Value": [np.nan] * n_methods,
                "Verdict": [""] * n_methods
            })

            if use_bf_table:
                self._tables.twoway_bf = table
            else:
                self._tables.twoway = table

        else:  # CUSTOM Hypothesis
            contrast = self._groups.contrast

            if self._groups.contrast_factor == 1:
                group1 = "+".join(
                    label for label, val in zip(self._labels.primary, contrast) if val >= 1
                )
                group2 = "+".join(
                    label for label, val in zip(self._labels.primary, contrast) if val <= -1
                )
            elif self._groups.contrast_factor == 2:
                group1 = "+".join(
                    label for label, val in zip(self._labels.secondary, contrast) if val >= 1
                )
                group2 = "+".join(
                    label for label, val in zip(self._labels.secondary, contrast) if val <= -1
                )
            else:
                raise ValueError("contrast_factor must be 1 or 2 for CUSTOM hypothesis.")

            self.hypothesis_LABEL = f"{group1} vs. {group2}"
            table = pd.DataFrame({
                f"{self.hypothesis_LABEL} Effect": method_col,
                "Test-Statistic": [np.nan] * n_methods,
                "P-Value": [np.nan] * n_methods,
                "Verdict": [""] * n_methods
            })

            if use_bf_table:
                self._tables.twoway_bf = table
            else:
                self._tables.twoway = table

        return pair_vec

    def _setup_time_bar(self, method):
        match method:
            case "L2-Bootstrap":
                match self.hypothesis:
                    case 'FAMILY':
                        desc = 'Calculating, Family-wise, Bootstrap L2 test'
                    case "PAIRWISE":
                        desc = f'Calculating, Pair-wise ({self._labels.hypothesis}), Bootstrap L2 test'
                    case "PRIMARY":
                        desc = f'Calculating, Primary Effect, Bootstrap L2 test'
                    case "SECONDARY":
                        desc = f'Calculating, Secondary Effect, Bootstrap L2 test'
                    case "INTERACTION":
                        desc = f'Calculating, Interaction Effect, Bootstrap L2 test'
                    case "CUSTOM":
                        desc = f'Calculating, Custom Contrast, Bootstrap L2 test'
                    case _:
                        raise ValueError(f'Unsupported Hypothesis: {self.hypothesis}')
            case "F-Bootstrap":
                match self.hypothesis:
                    case 'FAMILY':
                        desc ='Calculating, Family-wise, Bootstrap F-type test'
                    case "PAIRWISE":
                        desc = f'Calculating, Pair-wise ({self._labels.hypothesis}), Bootstrap F-type test'
                    case "PRIMARY":
                        desc = f'Calculating, Primary Effect, Bootstrap F-type test'
                    case "SECONDARY":
                        desc = f'Calculating, Secondary Effect, Bootstrap F-type test'
                    case "INTERACTION":
                        desc = f'Calculating, Interaction Effect, Bootstrap F-type test'
                    case "CUSTOM":
                        desc = f'Calculating, Custom Contrast, Bootstrap F-type test'
                    case _:
                        raise ValueError(f'Unsupported Hypothesis: {self.hypothesis}')
            case _:
                raise ValueError(f'Unsupported method for TQDM: {method}')

        return desc

    def _n_ii_generator(self):
        """
        Creates self.n_ii: a list of lists containing sample sizes for each
        (primary, secondary) combination — used for plotting in Two-Way ANOVA.
        """

        subgroup = cast(np.ndarray, self._groups.subgroup_indicator) #  it's only for static type checking.
        labels_primary = cast(list, self._labels.primary) #  it's only for static type checking.
        labels_secondary = cast(list, self._labels.secondary) #  it's only for static type checking.

        aflag = utils.aflag_maker(self.n_i)
        bflag = subgroup

        aflag_levels = np.unique(aflag)
        bflag_levels = np.unique(bflag)

        p = len(aflag_levels)  # primary levels
        q = len(bflag_levels)  # secondary levels

        p_cell = []
        error_messages = []

        for i in range(p):
            q_cell = [np.nan] * q
            for j in range(q):
                # Find indices where both flags match the (i, j) combination
                ij_flag = (aflag == aflag_levels[i]) & (bflag == bflag_levels[j])
                n_ij = np.sum(ij_flag)

                if n_ij == 0:
                    error_messages.append(
                        f"A missing combination of data occurs for Primary Label: {labels_primary[i]} and Secondary Label: {labels_secondary[j]}"
                    )

                q_cell[j] = n_ij
            p_cell.append(q_cell)

        if error_messages:
            for msg in error_messages:
                warnings.warn(msg)
            raise ValueError("Missing combinations detected. See warnings above.")

        self.n_ii = p_cell

    def _computeSSH_and_pairs(self, eta_i, eta_grand):
        pair_vec = []

        if self.hypothesis == "FAMILY":
            q = self._groups.k - 1  # rank of contrast matrix
            n_tests = 1
            SSH_k = np.zeros((self.n_domain_points, self._groups.k))

            for k in range(self._groups.k):
                SSH_k[:, k] = self.n_i[k] * (eta_i[:, k] - eta_grand) ** 2

            SSH_t = np.sum(SSH_k, axis=1)
            pair_vec = ["FAMILY"]
            C = None
            D = None

        elif self.hypothesis == "PAIRWISE":
            C = utils.construct_pairwise_contrast_matrix(self._groups.k)  # Expected shape: (n_tests, k_groups)
            n_tests = C.shape[0]

            if self._labels.group is None:
                raise ValueError('Group labels must be set')

            assert len(self._labels.group) == self._groups.k, "Each Group Must Have Exactly One Label Associated to it"

            c = np.zeros((1, self.n_domain_points))
            D = np.diag(1.0 / np.array(self.n_i))
            q = 1  # rank of PAIRWISE contrasts
            SSH_t = np.zeros((self.n_domain_points, n_tests))
            pair_vec = []

            for cc in range(n_tests):
                Ct = C[cc, :]  # shape: (k_groups,)
                part12 = Ct @ eta_i.T - c  # shape: (1, n_domain_points)
                rh_side = Ct @ D @ Ct.T

                if rh_side.ndim == 0:
                    SSH_t[:, cc] = (part12 ** 2).flatten() * 1/rh_side
                else:
                    SSH_t[:, cc] = (part12 ** 2).flatten() * np.linalg.inv(rh_side)



                # Label for the contrast
                indices = np.where(C[cc, :] != 0)[0]
                t1 = self._labels.group[indices[0]]
                t2 = self._labels.group[indices[1]]
                pair_vec.append(f"{t1} & {t2}")
        else:
            raise ValueError(f'Unknown Hypothesis: {self.hypothesis}')

        return HypothesisInfo(SSH_t, pair_vec, q, n_tests, C, D)

    def _compute_contrast_matrix(self):
        """
        Define the contrast matrix `C` and number of tests `n_tests`
        using pattern matching on `self.Hypothesis`.
        """
        match self.hypothesis:
            case "PAIRWISE":
                C = utils.construct_pairwise_contrast_matrix(self._groups.AB)
                n_tests = C.shape[0]  # AB_groups choose 2

            case "FAMILY":
                n_tests = 1
                C = np.hstack([
                    np.eye(self._groups.AB - 1),
                    -np.ones((self._groups.AB - 1, 1))
                ])

            case "INTERACTION" | "PRIMARY" | "SECONDARY":
                n_tests = 1
                C = None  # or np.array([]) if needed

            case _:  # Default or "CUSTOM"
                n_tests = 1
                C = self._groups.contrast

        return C, n_tests

    @staticmethod
    def _get_rng(seed):
        if isinstance(seed, np.random.Generator):
            return seed
        return np.random.Generator(np.random.PCG64(seed)) # np.random.PCG64(None) is valid

    @staticmethod
    def _cast_to_1D(arr):
        arr = np.asarray(arr)

        if not np.issubdtype(arr.dtype, np.number):
            raise ValueError("d_grid must contain numeric values.")

        if arr.ndim == 1:
            return arr
        elif arr.ndim == 2 and (arr.shape[0] == 1 or arr.shape[1] == 1):
            return np.ravel(arr)  # returns a view if possible
        else:
            raise ValueError(f"Input must be a 1D vector, or 2D row/column vector, but got array with shape {arr.shape}")

    @staticmethod
    def sanitize_contrast(c):
        if c is None:
            return np.empty((0,), float)  # length 0
        arr = np.asarray(c)
        if arr.ndim == 0:
            # if it’s literally the scalar None, make it empty; otherwise make it length-1
            if arr.dtype == object and arr.item() is None:
                return np.empty((0,), float)
            return arr.reshape(1,)
        return arr



import numpy as np
import pytest
from typing import Callable
from dataclasses import dataclass, asdict
from fanova import functionalANOVA
from utils import *

@dataclass
class LogisticParams:
    L: float = 7
    k: float = 0.75
    x_0: float = 10

# contains just length values for rbf kernel
@dataclass
class RBFParams:
    l1: float = 0.25
    l2: float = 2.5
    l3: float = 25


@pytest.fixture(scope="session")
def log_params() -> LogisticParams:
    return LogisticParams()

@pytest.fixture(scope="session")
def rbf_params() -> RBFParams:
    return RBFParams()

@pytest.fixture(scope="session")
def logistic_init(log_params: LogisticParams, rbf_params: RBFParams) -> dict:
    t_grid = np.arange(0,25,0.01)
    eta_t = logistic_function(log_params.L, log_params.k, log_params.x_0, t_grid)
    l_values: np.ndarray = np.array(list(asdict(rbf_params).values()))

    cov_matricies: list[np.ndarray] = []
    for l in l_values:
        kernel = lambda s,t: rbf_kernel(s,t,l)
        cov_matricies.append(covariance_matrix(kernel, t_grid))

    return {
        "t_grid": t_grid,
        "eta_t": eta_t,
        "cov_matricies": cov_matricies,
    }


class TestOneway:
    @pytest.mark.parametrize("cov_idx",[
        pytest.param(0,id="l=0.25"),
        pytest.param(1,id="l=2.5"),
        pytest.param(2,id="l=25"),
    ])
    def test_equal_means(self, logistic_init: dict,cov_idx: int) -> None:
        rng: np.random.Generator = init_rng(9000)
        n_reals: int = 25
        n_groups: int = 3

        cov: np.ndarray = logistic_init["cov_matricies"][cov_idx]

        groups: list[np.ndarray] = []
        for k in range(n_groups):
            data = generate_logistic_data(logistic_init["eta_t"], cov, n_reals, rng)
            groups.append(data)

        fanova: functionalANOVA = functionalANOVA(
            groups,
            d_grid=logistic_init["t_grid"],
            grid_bounds=(-np.inf,np.inf)
        )

        fanova.oneway(
            hypothesis="FAMILY",
            seed=9000
        )
        df = fanova.tables.oneway
        assert df is not None
        assert (df["P-Value"] > 0.05).all()

    @pytest.mark.parametrize("hypothesis_type", [
        "FAMILY",
        "PAIRWISE"
    ])
    def test_shifted_means(self, logistic_init: dict, hypothesis_type: str) -> None:
        rng: np.random.Generator = init_rng(9001)
        n_reals: int = 15
        n_groups: int = 3

        cov: np.ndarray = logistic_init["cov_matricies"][1]

        groups: list[np.ndarray] = []

        for g in range(n_groups):
            data: np.ndarray = generate_logistic_data(logistic_init["eta_t"], cov, n_reals, rng)
            if g == 0:
                data += rng.integers(1,6)
            groups.append(data)

        fanova: functionalANOVA = functionalANOVA(
            groups,
            d_grid=logistic_init["t_grid"],
            grid_bounds=(-np.inf,np.inf)
        )

        fanova.oneway(
            hypothesis=hypothesis_type,
            seed=9001
        )

        assert fanova.tables.oneway is not None
        if hypothesis_type == "FAMILY":
            df = fanova.tables.oneway
            assert df is not None
            assert np.all(df["P-Value"] < 0.05)
        else: # pairwise
            assert np.all(fanova.tables.oneway.iloc[0,1:] < 0.05)  # 0 & 1
            assert np.all(fanova.tables.oneway.iloc[1,1:] < 0.05)  # 0 & 2
            assert np.all(fanova.tables.oneway.iloc[2,1:] > 0.05)  # 1 & 2

class TestOnewayBF:
    def test_equal_means(self, logistic_init: dict) -> None:
        rng: np.random.Generator = init_rng(9001)
        n_reals: int = 30
        n_groups: int = 3

        groups: list[np.ndarray] = []
        for k in range(n_groups):
            data = generate_logistic_data(logistic_init["eta_t"], logistic_init["cov_matricies"][k], n_reals, rng)
            groups.append(data)

        fanova: functionalANOVA = functionalANOVA(
            groups,
            d_grid=logistic_init["t_grid"],
            grid_bounds=(-np.inf, np.inf)
        )

        fanova.oneway_bf(
            hypothesis="FAMILY",
            seed=9001
        )
        

        df = fanova.tables.oneway_bf
        assert df is not None
        assert np.all(df["P-Value"] > 0.05)

    @pytest.mark.parametrize("hypothesis_type", [
        "FAMILY",
        "PAIRWISE"
    ])
    def test_shifted_means(self, logistic_init: dict, hypothesis_type: str) -> None:
        rng: np.random.Generator = init_rng(9001)
        n_reals: int = 30
        n_groups: int = 3

        groups: list[np.ndarray] = []
        for k in range(n_groups):
            data = generate_logistic_data(logistic_init["eta_t"], logistic_init["cov_matricies"][k], n_reals, rng)
            if k == 2:
                data += 1
            groups.append(data)

        fanova: functionalANOVA = functionalANOVA(
            groups,
            d_grid=logistic_init["t_grid"],
            grid_bounds=(-np.inf, np.inf)
        )

        fanova.oneway_bf(
            hypothesis=hypothesis_type,
            seed=9001
        )

        if hypothesis_type == "FAMILY":
            df = fanova.tables.oneway_bf
            assert df is not None
            assert np.all(df["P-Value"] < 0.05)
        else: # pairwise
            table = fanova.tables.oneway_bf
            assert table is not None
            assert np.all(table.iloc[0,1:] > 0.05)  # 0 & 1
            assert np.all(table.iloc[1,1:] < 0.05)  # 0 & 2
            assert np.all(table.iloc[2,1:] < 0.05)  # 1 & 2



import pandas as pd
import numpy as np
import pytest
from importlib import resources
from fanova import functionalANOVA


@pytest.fixture(scope="session")
def gait_data() -> dict[str, np.ndarray]:
    # Load packaged dataset
    with resources.files("fanova.datasets").joinpath("gait_data.csv").open("r") as f:
        df: pd.DataFrame = pd.read_csv(f)

    t_grid: np.ndarray = df["t"].to_numpy()
    group1: np.ndarray = df.filter(like="group1").to_numpy()
    group2: np.ndarray = df.filter(like="group2").to_numpy()
    group3: np.ndarray = df.filter(like="group3").to_numpy()

    return {
        "t_grid": t_grid,
        "group1": group1,
        "group2": group2,
        "group3": group3
    }


@pytest.fixture(scope="session")
def r_verif() -> dict[str, float]:
    return {
        "L2_stat": 2637.128205128205,
        "F_stat": 1.4682175466420953,
        "L2_Naive_pval": 0.2106562,
        "L2_BiasReduced_pval": 0.1957646,
        "F_Naive_pval": 0.2226683,
        "F_BiasReduced_pval": 0.2198691,
    }


class TestRVerification:

    @pytest.fixture(autouse=True)
    def _run_fanova(self, gait_data: dict[str, np.ndarray]) -> None:
        groups = [
            gait_data["group1"],
            gait_data["group2"],
            gait_data["group3"],
        ]

        self.fanova = functionalANOVA(
            groups,
            d_grid=gait_data["t_grid"],
            grid_bounds=(-np.inf, np.inf),
            group_labels=["1", "2", "3"],
        )

        self.fanova.oneway(hypothesis="FAMILY")

    def test_l2_stat(self, r_verif: dict[str, float]) -> None:
        table = self.fanova.tables.oneway
        assert table is not None
        
        l2_stat = table.loc[
            table["Family-Wise Method"] == "L2-Naive",
            "Test-Statistic",
        ].iloc[0]

        np.testing.assert_allclose(l2_stat, r_verif["L2_stat"], rtol=1e-2)

    def test_f_stat(self, r_verif: dict[str, float]) -> None:
        table = self.fanova.tables.oneway
        assert table is not None
        f_stat = table.loc[
            table["Family-Wise Method"] == "F-Naive",
            "Test-Statistic",
        ].iloc[0]

        np.testing.assert_allclose(f_stat, r_verif["F_stat"], rtol=1e-2)

    def test_l2_pvals(self, r_verif: dict[str, float]) -> None:
        table = self.fanova.tables.oneway
        assert table is not None
        for method, key in [
            ("L2-Naive", "L2_Naive_pval"),
            ("L2-BiasReduced", "L2_BiasReduced_pval"),
        ]:
            pval = table.loc[
                table["Family-Wise Method"] == method,
                "P-Value",
            ].iloc[0]

            np.testing.assert_allclose(pval, r_verif[key], rtol=1e-2)

    def test_f_pvals(self, r_verif: dict[str, float]) -> None:
        table = self.fanova.tables.oneway
        assert table is not None
        
        for method, key in [
            ("F-Naive", "F_Naive_pval"),
            ("F-BiasReduced", "F_BiasReduced_pval"),
        ]:
            pval = table.loc[
                table["Family-Wise Method"] == method,
                "P-Value",
            ].iloc[0]

            np.testing.assert_allclose(pval, r_verif[key], rtol=1e-2)


import numpy as np
from scipy.stats import chi2, gaussian_kde
from fanova import utils

# TODO: Needs work
def k_group_cov_pairwise(self, method, y1, y2, seed=None):

    rng = self._get_rng(seed)

    n1, L = y1.shape
    n2, _ = y2.shape
    gsize = [n1, n2]

    Sigma1 = np.cov(y1, rowvar=False)
    Sigma2 = np.cov(y2, rowvar=False)
    N = n1 + n2
    # Sigma = ((n1-1)*Sigma1 + (n2-1)*Sigma2) / (N-2)
    stat = (n1 - 1) * (n2 - 1) / (N - 2) * np.trace((Sigma1 - Sigma2) @ (Sigma1 - Sigma2))  # Corrected

    # TESTING
    vmu = []
    V = []
    n_i = [n1, n2]
    temp_data = [y1, y2]

    for ii in range(2):
        yyi = temp_data[ii]
        mui = np.mean(yyi, axis=0)
        Vi = yyi - np.ones((n_i[ii], 1)) @ mui.reshape(1, -1)
        vmu.append(mui)
        V.append(Vi)

    vmu = np.array(vmu)
    V = np.vstack(V)

    m = self.n_domain_points

    if N > m:
        Sigma = (V.T @ V) / (N - 2)  # mxm
    else:
        Sigma = (V @ V.T) / (N - 2)  # NxN

    if method == "L2-Simul":
        q = 1

        v_1j = y1.T - np.mean(y1, axis=0, keepdims=True).T
        v_2j = y2.T - np.mean(y2, axis=0, keepdims=True).T

        n_array = [n1, n2]
        v_array = [v_1j, v_2j]
        LHS = 0

        ts = set_up_time_bar('Calculating Simulated Null Distribution', sum(n_array))

        for ii in range(2):
            n_i = n_array[ii]
            V_group = v_array[ii]
            for jj in range(n_i):
                v_ij = V_group[:, jj]
                LHS += np.outer(v_ij, v_ij) @ np.outer(v_ij, v_ij)
                ts.progress()

        ts.delete()

        LHS = LHS / N

        if LHS.shape == Sigma.shape:
            omega_hat = LHS - (Sigma @ Sigma)  # matrix multiplication
        else:
            SigmaLarge = ((n1 - 1) * np.cov(y1, rowvar=False) + (n2 - 1) * np.cov(y2, rowvar=False)) / (N - 2)
            omega_hat = LHS - (SigmaLarge @ SigmaLarge)  # matrix multiplication

        eig_gamma_hat = np.real(np.linalg.eigvals(omega_hat))
        eig_gamma_hat = eig_gamma_hat[eig_gamma_hat > 0]

        T_null = utils.chi_sq_mixture(q, eig_gamma_hat, self.n_simul, rng)
        kde = gaussian_kde(T_null)
        pvalue = 1 - kde.integrate_box_1d(-np.inf, stat)
        pvalue = max(0,min(1,pvalue))

    elif method == "L2-BiasReduced":  # Bias Reduced
        A = np.trace(Sigma @ Sigma) + np.trace(Sigma)**2
        B = 2 * np.trace(np.linalg.matrix_power(Sigma, 4)) + 2 * np.trace(Sigma @ Sigma)**2

        alpha = (N - 2)**2 / (N * (N - 3)) * (B - A**2 / (N - 2)) / A
        df = (1 + 1 / (N - 2)) * (A**2 - 2 * B / (N - 1)) / (B - A**2 / (N - 2))

        pvalue = 1 - chi2.cdf(stat / alpha, df)

    elif method == "L2-Naive":  # Naive Method
        an = np.trace(Sigma)
        bn = np.trace(Sigma @ Sigma)
        cn = np.trace(np.linalg.matrix_power(Sigma, 3))
        dn = np.trace(np.linalg.matrix_power(Sigma, 4))

        An = an
        Bn = (N - 2)**2 / (N * (N - 3)) * (bn - An**2 / (N - 2))
        Cn = (cn - 3 / (N - 2) * Bn * An) / (1 + 3 / (N - 2))
        Dn = (dn - 6 / (N - 2) * Cn * An) / (1 + 6 / (N - 2))
        A = Bn + An**2
        B = 2 * Dn + 2 * Bn**2
        alpha = B / A
        df = A**2 / B

        pvalue = 1 - chi2.cdf(stat / alpha, df)

    elif method == "Bootstrap-Test":  # Bootstrap test
        ts = set_up_time_bar('Running Bootstrap Test', self.n_boot)
        vstat = np.zeros(self.n_boot)
        k = 2

        def bootstrap_iteration(ii):
            flag1 = rng.choice(n1, n1, replace=True)
            flag2 = rng.choice(n2, n2, replace=True)

            yy1 = y1[flag1, :]
            yy2 = y2[flag2, :]

            # Computing p_s : Bootstrap Pooled Covariance
            mu1 = np.mean(yy1, axis=0)
            R1 = yy1 - np.ones((n1, 1)) @ mu1.reshape(1, -1)

            mu2 = np.mean(yy2, axis=0)
            R2 = yy2 - np.ones((n2, 1)) @ mu2.reshape(1, -1)

            R = np.vstack([R1, R2])

            if N > m:
                p_s = (R.T @ R) / (N - k)  # p x p pooled covariance matrix
            else:
                p_s = (R @ R.T) / (N - k)

            stat0 = 0
            nni = 0
            for i in range(k):
                ni = gsize[i]
                flag = np.arange(nni, nni + ni)
                Ri = R[flag, :]

                # ith group's covariance
                if N > m:
                    p_si = (Ri.T @ Ri) / (ni - 1)  # Vi: ni x p
                    temp = np.trace(np.linalg.matrix_power(p_si - p_s, 2))
                else:
                    p_si = (Ri @ Ri.T) / (ni - 1)
                    temp = (np.trace(np.linalg.matrix_power(p_si, 2)) -
                           2 * np.trace(((Ri @ R.T) @ R) @ Ri.T) / (N - k) / (ni - 1) +
                           np.trace(np.linalg.matrix_power(p_s, 2)))

                stat0 += (ni - 1) * temp
                nni += ni

            return stat0

        # Bootstrap iterations
        for ii in range(self.n_boot):
            vstat[ii] = bootstrap_iteration(ii)
            ts.progress()

        ts.delete()
        pvalue = np.mean(vstat > stat)

    elif method == "Permutation-Test":  # Permutation test
        ts = set_up_time_bar('Running Permutation Test', self.N_permutations)
        vstat = np.zeros(self.N_permutations)
        k = 2
        Y = np.vstack([y1, y2])

        def permutation_iteration(ii):
            flag = rng.permutation(N)
            Y_perm = Y[flag, :]
            yy1 = Y_perm[:n1, :]
            yy2 = Y_perm[n1:, :]

            mu1 = np.mean(yy1, axis=0)
            R1 = yy1 - np.ones((n1, 1)) @ mu1.reshape(1, -1)

            mu2 = np.mean(yy2, axis=0)
            R2 = yy2 - np.ones((n2, 1)) @ mu2.reshape(1, -1)

            R = np.vstack([R1, R2])

            if N > m:
                p_s = (R.T @ R) / (N - k)
            else:
                p_s = (R @ R.T) / (N - k)

            stat0 = 0
            nni = 0
            for i in range(k):
                ni = gsize[i]
                flag = np.arange(nni, nni + ni)
                Ri = R[flag, :]

                if N > m:
                    p_si = (Ri.T @ Ri) / (ni - 1)
                    temp = np.trace(np.linalg.matrix_power(p_si - p_s, 2))
                else:
                    p_si = (Ri @ Ri.T) / (ni - 1)
                    temp = (np.trace(np.linalg.matrix_power(p_si, 2)) -
                           2 * np.trace(((Ri @ R.T) @ R) @ Ri.T) / (N - k) / (ni - 1) +
                           np.trace(np.linalg.matrix_power(p_s, 2)))

                stat0 += (ni - 1) * temp
                nni += ni

            return stat0

        # Permutation iterations
        for ii in range(self.N_permutations):
            vstat[ii] = permutation_iteration(ii)
            ts.progress()

        ts.delete()
        pvalue = np.mean(vstat > stat)

    return pvalue

# TODO: Needs work
def k_group_cov(self, method, stat, Sigma, V, seed=None):

    rng = self._get_rng(seed)

    gsize = self.n_i
    N = self.N
    k = self.k_groups
    p = self.n_domain_points

    if method == "L2-Simul":
        q = self.k_groups - 1
        v_array = []
        for K in range(self.k_groups):
            v_array.append(self.data[K] - np.mean(self.data[K], axis=1, keepdims=True))

        n_array = self.n_i
        LHS = 0

        for ii in range(self.k_groups):
            n_i = n_array[ii]
            V_i = v_array[ii]
            for jj in range(n_i):
                v_ij = V_i[:, jj]
                LHS += np.outer(v_ij, v_ij) * np.outer(v_ij, v_ij)

        LHS /= N

        if LHS.shape == Sigma.shape:
            omega_hat = LHS - (Sigma @ Sigma)
        else:
            vmu = []
            V_large = []

            for ii in range(self.k_groups):
                yyi = self.data[ii].T
                mui = np.mean(yyi, axis=0)

                Vi = yyi - np.ones((self.n_i[ii], 1)) * mui
                vmu.append(mui)
                V_large.append(Vi)

            V_large = np.vstack(V_large)
            SigmaLarge = (V_large.T @ V_large) / (self.N - self.k_groups)
            omega_hat = LHS - (SigmaLarge @ SigmaLarge)

        eig_gamma_hat = np.real(np.linalg.eigvals(omega_hat))
        eig_gamma_hat = eig_gamma_hat[eig_gamma_hat > 0]

        T_null = utils.chi_sq_mixture(q, eig_gamma_hat, self.n_simul, rng)
        kde = gaussian_kde(T_null)
        pvalue = 1 - kde.integrate_box_1d(-np.inf, stat)
        pvalue = max(0,min(1,pvalue))

    elif method == "L2-Naive":
        an = np.trace(Sigma)

        # np.lingalg.matrix_power works better than dot products
        bn = np.trace(np.linalg.matrix_power(Sigma, 2))
        cn = np.trace(np.linalg.matrix_power(Sigma, 3))
        dn = np.trace(np.linalg.matrix_power(Sigma, 4))

        Bn = (N - 2)**2 / (N * (N - 3)) * (bn - an**2 / (N - 2))
        Cn = (cn - (3 / (N - 2)) * Bn * an) / (1 + 3 / (N - 2))
        Dn = (dn - (6 / (N - 2)) * Cn * an) / (1 + 6 / (N - 2))

        A_corr = Bn + an**2
        B_corr = 2 * Dn + 2 * Bn**2

        alpha = B_corr / A_corr
        df = (k - 1) * (A_corr**2) / B_corr

        pvalue = 1 - chi2.cdf(stat / alpha, df)

    elif method == "L2-BiasReduced":
        A0 = np.trace(Sigma @ Sigma) + np.trace(Sigma)**2

        B0 = 2 * np.trace(np.linalg.matrix_power(Sigma, 4)) + 2 * (np.trace(Sigma @ Sigma))**2


        alpha = ((N - k)**2 / (N * (N - k - 1)) * (B0 - A0**2 / (N - k))) / A0
        df = (1 + 1 / (N - k)) * (A0**2 - 2 * B0 / (N - k + 1)) / (B0 - A0**2 / (N - k))
        df = (k - 1) * df

        pvalue = 1 - chi2.cdf(stat / alpha, df)

    elif method == "Bootstrap-Test":
        aflag = utils.aflag_maker(gsize)
        aflag0 = np.unique(aflag)
        vstat = np.zeros(self.n_boot)

        for ii in range(self.n_boot):
            flag = rng.choice(N, N, replace=True)
            py = V[flag, :]

            R = []

            for i in range(k):
                iflag = (aflag == aflag0[i])
                yyi = py[iflag, :]
                ni = gsize[i]
                mui = np.mean(yyi, axis=0)
                Ri = yyi - np.ones((ni, 1)) * mui
                R.append(Ri)

            R = np.vstack(R)

            if N > p:
                p_s = (R.T @ R) / (N - k)
            else:
                p_s = (R @ R.T) / (N - k)

            stat0 = 0
            nni = 0
            for i in range(k):
                ni = gsize[i]
                flag = np.arange(nni, nni + ni)
                Ri = R[flag, :]

                if N > p:
                    p_si = (Ri.T @ Ri) / (ni - 1)
                    temp = np.trace(np.linalg.matrix_power(p_si - p_s, 2))
                else:
                    p_si = (Ri @ Ri.T) / (ni - 1)
                    temp = np.trace(np.linalg.matrix_power(p_si, 2)) - 2 * np.trace(((Ri @ R.T) @ R) @ Ri.T) / (N - k) / (ni - 1) + np.trace(np.linalg.matrix_power(p_s, 2))

                stat0 += (ni - 1) * temp
                nni += ni
            vstat[ii] = stat0
        pvalue = 1 - np.sum(vstat < stat) / self.n_boot

    elif method == "Permutation-Test":
        aflag = utils.aflag_maker(gsize)
        aflag0 = np.unique(aflag)

        vstat = np.zeros(self.N_permutations)
        for ii in range(self.N_permutations):
            flag = rng.permutation(N)
            py = V[flag, :]

            R = []
            for i in range(k):
                iflag = (aflag == aflag0[i])
                yyi = py[iflag, :]
                ni = gsize[i]
                mui = np.mean(yyi, axis=0)
                Ri = yyi - np.ones((ni, 1)) * mui
                R.append(Ri)

            R = np.vstack(R)

            if N > p:
                p_s = (R.T @ R) / (N - k)
            else:
                p_s = (R @ R.T) / (N - k)

            stat0 = 0
            nni = 0
            for i in range(k):
                ni = gsize[i]
                flag = np.arange(nni, nni + ni)
                Ri = R[flag, :]

                if N > p:
                    p_si = (Ri.T @ Ri) / (ni - 1)
                    temp = np.trace(np.linalg.matrix_power(p_si - p_s, 2))

                else:
                    p_si = (Ri @ Ri.T) / (ni - 1)
                    temp = np.trace(np.linalg.matrix_power(p_si, 2)) - 2 * np.trace(((Ri @ R.T) @ R) @ Ri.T) / (N - k) / (ni - 1) + np.trace(np.linalg.matrix_power(p_s, 2))

                stat0 += (ni - 1) * temp
                nni += ni
            vstat[ii] = stat0

        pvalue = 1 - np.sum(vstat < stat) / self.N_permutations

    return pvalue

# TODO: Needs work
def two_group_cov(self, method, y1, y2, seed=None):

    rng = self._get_rng(seed)

    n1, L = y1.shape
    n2, _ = y2.shape

    Sigma1 = np.cov(y1, rowvar=False)
    Sigma2 = np.cov(y2, rowvar=False)
    N = n1 + n2

    Sigma = ((n1 - 1) * Sigma1 + (n2 - 1) * Sigma2) / (N - 2)
    stat = (n1 - 1) * (n2 - 1) / (N - 2) * np.trace((Sigma1 - Sigma2) @ (Sigma1 - Sigma2))

    if method == "L2-Simul":
        q = 1
        v_1j = y1.T - np.mean(y1, axis=0, keepdims=True).T
        v_2j = y2.T - np.mean(y2, axis=0, keepdims=True).T

        n_array = [n1, n2]
        v_array = [v_1j, v_2j]
        LHS = 0

        for ii in range(2):
            n_i = n_array[ii]
            V = v_array[ii]
            for jj in range(n_i):
                v_ij = V[:, jj]
                LHS += np.outer(v_ij, v_ij) @ np.outer(v_ij, v_ij)

        LHS /= N

        if LHS.shape == Sigma.shape:
            omega_hat = LHS - (Sigma @ Sigma)
        else:
            SigmaLarge = ((n1 - 1) * np.cov(y1, rowvar=False) + (n2 - 1) * np.cov(y2, rowvar=False)) / (N - 2)
            omega_hat = LHS - (SigmaLarge @ SigmaLarge)

        eig_gamma_hat = np.real(np.linalg.eigvals(omega_hat))
        eig_gamma_hat = eig_gamma_hat[eig_gamma_hat > 0]

        T_null = utils.chi_sq_mixture(q, eig_gamma_hat, self.n_simul, rng)

        kde = gaussian_kde(T_null)
        pvalue = 1 - kde.integrate_box_1d(-np.inf, stat)
        pvalue = max(0,min(1,pvalue))

    elif method == "L2-BiasReduced":
        A = np.trace(Sigma @ Sigma) + np.trace(Sigma)**2
        B = 2 * np.trace(np.linalg.matrix_power(Sigma, 4)) + 2 * np.trace(Sigma @ Sigma)**2

        alpha = (N - 2)**2 / (N * (N - 3)) * (B - A**2 / (N - 2)) / A
        df = (1 + 1 / (N - 2)) * (A**2 - 2 * B / (N - 1)) / (B - A**2 / (N - 2))

        pvalue = 1 - chi2.cdf(stat / alpha, df)

    elif method == "L2-Naive":
        an = np.trace(Sigma)

        bn = np.trace(Sigma @ Sigma)
        cn = np.trace(np.linalg.matrix_power(Sigma, 3))
        dn = np.trace(np.linalg.matrix_power(Sigma, 4))

        An = an
        Bn = (N - 2)**2 / (N * (N - 3)) * (bn - An**2 / (N - 2))
        Cn = (cn - 3 / (N - 2) * Bn * An) / (1 + 3 / (N - 2))
        Dn = (dn - 6 / (N - 2) * Cn * An) / (1 + 6 / (N - 2))

        A = Bn + An**2
        B = 2 * Dn + 2 * Bn**2
        alpha = B / A
        df = A**2 / B
        pvalue = 1 - chi2.cdf(stat / alpha, df)

    elif method == "Bootstrap-Test":
        vstat = np.zeros(self.n_boot)

        for ii in range(self.n_boot):
            flag1 = rng.choice(n1, n1, replace=True)
            flag2 = rng.choice(n2, n2, replace=True)

            yy1 = y1[flag1, :]
            yy2 = y2[flag2, :]

            S1 = np.cov(yy1, rowvar=False)
            S2 = np.cov(yy2, rowvar=False)

            stat0 = (n1 - 1) * (n2 - 1) / (N - 2) * np.trace(np.linalg.matrix_power((S1 - S2) - (Sigma1 - Sigma2), 2))
            vstat[ii] = stat0
        pvalue = np.mean(vstat > stat)

    elif method == "Permutation-Test":
        vstat = np.zeros(self.N_permutations)
        y1_centered = y1 - np.mean(y1, axis=0)
        y2_centered = y2 - np.mean(y2, axis=0)
        yy = np.vstack((y1_centered, y2_centered))

        for ii in range(self.N_permutations):
            flag = rng.permutation(N)

            yy1 = yy[flag[:n1], :]
            yy2 = yy[flag[n1:], :]

            S1 = np.cov(yy1, rowvar=False)
            S2 = np.cov(yy2, rowvar=False)

            stat0 = (n1 - 1) * (n2 - 1) / (N - 2) * np.trace(np.linalg.matrix_power(S1 - S2, 2))
            vstat[ii] = stat0

        pvalue = np.mean(vstat > stat)

    return pvalue



from importlib import resources
from scipy.io import loadmat
import pandas as pd


def load_example_csv() -> pd.DataFrame:
    with resources.files("fanova.datasets").joinpath("gait_data.csv").open("rb") as f:
        df = pd.read_csv(f, encoding="utf-8-sig")
    return df

def load_example_mat():
    with resources.files("fanova.datasets").joinpath("example_data.mat").open("rb") as f:
        return loadmat(f)

def load_example_mat2():
    with resources.files("fanova.datasets").joinpath("example_data_2.mat").open("rb") as f:
        return loadmat(f)


import numpy as np
from scipy import stats
from scipy.stats import chi2, f
from scipy.linalg import inv
import pandas as pd
from tqdm import tqdm
from fanova import utils
from fanova.methods.oneway import run_onewayBF

# TODO: Check Pairwise hypothesis
def run_twoway(self, method, data, contrast, rng):

    N = self.N
    ddim = self.n_domain_points
    aflag = utils.aflag_maker(self.n_i)
    bflag = self._groups.subgroup_indicator
    aflag0 = np.unique(aflag)
    p = len(aflag0)

    yy = data

    bflag0 = np.unique(bflag)

    q = len(bflag0)
    k = (p * q)

    Hp = 0
    Hq = 0
    Ap = 0
    Aq = 0
    u = v = None  # makes them always "defined"

    gsize = []
    vmu = []
    V = []
    p_cell = []
    error_string = []

    for i in range(p):
        q_cell = []
        for j in range(q):
            ijflag = (aflag == aflag0[i]) & (bflag == bflag0[j])
            ni = np.sum(ijflag)

            if ni == 0:
                error_string.append(f"A missing combination of data occurs for Primary Label {self._labels.primary[i]} and Secondary Label {self._labels.secondary[j]}")

            gsize.append(ni)
            yyi = yy[ijflag, :]
            cell_mui = np.mean(yyi, axis=0)
            vmu.append(cell_mui)
            V.extend(yyi - cell_mui)
            q_cell.append(ni)
        p_cell.append(q_cell)

    if error_string:
        for i in range(len(error_string)):
            print(f"warning: {i}")

        raise ValueError("'Missing Combinations Listed Above'")

    self.n_ii = p_cell
    n_ii = np.array([i for j in p_cell for i in j])

    gsize = np.array(gsize)

    vmu = np.array(vmu)
    V = np.array(V)

    if self.hypothesis not in ["PAIRWISE", "FAMILY"]:
        # Define u and v based on weights
        if self.weights.upper() == "UNIFORM":
            u = np.ones(p) / p
            v = np.ones(q) / q

        elif self.weights.upper() == "PROPORTIONAL":
            matsize = np.array(p_cell)
            N_total = np.sum(gsize)

            u = np.sum(matsize, axis=1) / N_total
            v = np.sum(matsize, axis=0) / N_total

        else:
            raise ValueError(f"Unsupported weight type: {self.weights}, must either be 'UNIFORM' or 'PROPORTIONAL'")


        Hp = np.hstack([np.eye(p - 1), -np.ones((p - 1, 1))])
        Hq = np.hstack([np.eye(q - 1), -np.ones((q - 1, 1))])

        Ap = np.eye(p) - np.outer(np.ones(p), u)
        Aq = np.eye(q) - np.outer(np.ones(q), v)


    if contrast.ndim == 1:
        contrast = contrast[None, :]

    # Handle hypothesis-specific contrast construction
    match self.hypothesis:
        case "PAIRWISE" | "FAMILY":
                r = contrast.shape[0]
        case "INTERACTION":
            contrast = np.kron(Hp, Hq) @ np.kron(Ap, Aq)
            r = (p - 1) * (q - 1)

        case "PRIMARY":
            assert v is not None, "v must be defined before using reshape"
            contrast = Hp @ np.kron(Ap, v)
            r = p - 1

        case "SECONDARY":
            assert u is not None, "v must be defined before using reshape"
            contrast = Hq @ np.kron(u.reshape(1, -1), Aq)
            r = q - 1

        case "CUSTOM" | _:
            if self._groups.contrast_factor == 1:
                assert v is not None, "v must be defined before using reshape"
                contrast = contrast @ np.kron(Ap, v.reshape(1, -1))
            elif self._groups.contrast_factor == 2:
                assert u is not None, "v must be defined before using reshape"
                contrast = contrast @ np.kron(u.reshape(1, -1), Aq)
            else:
                raise ValueError("Invalid contrast_factor: must be 1 or 2")

            r = contrast.shape[0]

    W = inv(contrast @ np.diag(1/gsize) @ contrast.T)
    SSH = np.diag((contrast @ vmu).T @ W @ (contrast @ vmu))
    SSH0 = np.sum(SSH)

    SSE = np.diag(V.T @ V)
    SSE0 = np.sum(SSE)

    pool_coeff = N-k

    if N > ddim:
        Pooled_COVAR = (V.T @ V)/pool_coeff
    else:
        Pooled_COVAR = (V @ V.T)/pool_coeff

    A = np.trace(Pooled_COVAR)
    B = np.trace(Pooled_COVAR @ Pooled_COVAR)

    A2 = 0
    B2 = 0

    if "Naive" in method:
        A2 = A**2
        B2 = B

    if "BiasReduced" in method:
        pool_coeff = N-k
        # just used pool_coef instead of N-k explicitly
        A2 = (pool_coeff) * (pool_coeff+1) / (pool_coeff-1) / (pool_coeff+2) * (A**2-2*B/(pool_coeff+1))
        B2 = (pool_coeff)**2 / (pool_coeff-1)/(pool_coeff+2) * (B-A**2/(pool_coeff))


    pvalue = np.nan
    stat = np.nan
    match method:
        case "L2-Simul":
            stat = SSH0
            eig_gamma_hat = np.linalg.eigvalsh(Pooled_COVAR)
            eig_gamma_hat = eig_gamma_hat[eig_gamma_hat > 0]

            SSH_null = utils.chi_sq_mixture(r, eig_gamma_hat, self.n_simul, self.rng)
            SSH_NullFitted = stats.gaussian_kde(SSH_null)
            pvalue = 1 - SSH_NullFitted.integrate_box_1d(-np.inf, stat)
            pvalue = max(0,min(1,pvalue))

        case "L2-Naive" | "L2-BiasReduced":
            stat = SSH0
            beta = B2/A
            kappa = A2/B2
            pvalue = 1 - chi2.cdf(stat/beta, r * kappa)

        case "F-Naive" | "F-BiasReduced":
            stat = SSH0 / SSE0 * (N-k) / r
            kappa = A2/B2
            pvalue = 1 - f.cdf(stat, r * kappa, (N-k) * kappa)

        case "F-Simul":
            stat = SSH0 / SSE0 * (N-k) / r
            eig_gamma_hat = np.linalg.eigvalsh(Pooled_COVAR)
            eig_gamma_hat = eig_gamma_hat[eig_gamma_hat > 0]

            SSH_null = utils.chi_sq_mixture(r, eig_gamma_hat, self.n_simul, self.rng)
            SSE_null = utils.chi_sq_mixture(N-k, eig_gamma_hat, self.n_simul, self.rng)

            ratio = (N-k) / r

            F_null = SSH_null / SSE_null * ratio
            F_NullFitted = stats.gaussian_kde(F_null)
            pvalue = 1 - F_NullFitted.integrate_box_1d(-np.inf, stat)
            pvalue = max(0,min(1,pvalue))

        case "L2-Bootstrap":
                stat = SSH0
                Bstat = np.zeros(self.n_boot)
                W = inv(contrast @ np.diag(1/gsize) @ contrast.T)
                Bmu = np.empty((p * q, yy.shape[1]))
                for b in tqdm(range(self.n_boot), desc=self._setup_time_bar(method)):
                    counter = 0

                    for i in range(p):
                        for j in range(q):
                            ijflag = (aflag==aflag0[i]) & (bflag == bflag0[j])
                            ni = n_ii[counter]
                            yi = yy[ijflag,:]
                            # Bootflag = np.random.choice(ni,ni,replace=True)
                            Bootflag = rng.choice(ni,ni,replace=True)

                            Byi = yi[Bootflag,:]
                            Bmui = np.mean(Byi, axis=0)
                            Bmu[counter] = Bmui
                            counter += 1

                    diff = Bmu - vmu
                    proj = contrast @ diff
                    SSH = np.einsum('ij,ij->j', proj.T @ W, proj.T)
                    #  SSH = np.diag(((Bmu - vmu).T @ contrast.T) @ W @ (contrast @ (Bmu - vmu)))
                    Bstat[b] = np.sum(SSH)

                pvalue = np.mean(Bstat>stat)

        case "F-Bootstrap":
            stat = SSH0 / SSE0 * (N-k) / r
            ratio = (N-k)/r
            Bstat = np.zeros(self.n_boot)
            W = np.linalg.inv(contrast @ np.diag(1 / gsize) @ contrast.T)
            ratio = (N - k) / r
            Bstat = np.zeros(self.n_boot)
            Bmu = np.empty((p * q, yy.shape[1]))

            for b in tqdm(range(self.n_boot), desc=self._setup_time_bar(method)):
                V = []
                counter = 0

                for i in range(p):
                    for j in range(q):
                        ijflag = (aflag == aflag0[i]) & (bflag == bflag0[j])
                        ni = n_ii[counter]

                        yi = yy[ijflag, :]
                        # Bootflag = np.random.randint(0, ni, size=ni)
                        Bootflag = rng.integers(0, ni, size=ni)
                        Byi = yi[Bootflag, :]
                        Bmui = np.mean(Byi, axis=0)
                        Bmu[counter] = Bmui

                        V.append(Byi - Bmui)  # list of residuals

                        counter += 1

                V = np.vstack(V)  # concatenate only once
                diff = Bmu - vmu
                proj = contrast @ diff
                SSH = np.einsum('ij,ij->j', proj.T @ W, proj.T)
                SSE = np.sum(V ** 2, axis=0)  # faster than np.diag(V.T @ V)

                Bstat[b] = np.sum(SSH) / np.sum(SSE) * ratio

            pvalue = np.mean(Bstat > stat)

    return pvalue, stat

# TODO: Needs work
def run_twowayBF(self, method, data, contrast, c):

    bflag = self._groups.subgroup_indicator
    N = self.N
    aflag = utils.aflag_maker(self.n_i)
    dim = self.n_domain_points

    aflag0 = np.unique(aflag)
    p = len(aflag0)

    bflag0 = np.unique(bflag)
    q = len(bflag0)
    u = v = None  # makes them always "defined"

    gsize = np.zeros((p,q))
    yy = []

    p_cell = []

    ij = 0

    for i in range(p):
        q_cell = []

        for j in range(q):
            ij += 1
            ijflag = (aflag == aflag0[i]) & (bflag == bflag0[j])
            nij = np.sum(ijflag)
            gsize[i,j] = nij

            yij = data[ijflag,:]
            yy.append(np.column_stack([ij * np.ones(nij), yij]))
            q_cell.append(nij)
        p_cell.append(q_cell)

    self.n_ii = p_cell


    yy = np.vstack(yy)

    if self.weights.upper() == "UNIFORM":
        u = np.ones(p) / p
        v = np.ones(q) / q

    elif self.weights.upper() == "PROPORTIONAL":
        u = np.sum(gsize, axis=1) / N
        v = np.sum(gsize, axis=0) / N
    else:
        raise ValueError(f"Unsupported weight type: {self.weights}, must either be 'UNIFORM' or 'PROPORTIONAL'")


    Ap = np.eye(p) - np.outer(np.ones(p), u)
    Aq = np.eye(q) - np.outer(np.ones(q), v)

    if contrast is None or len(contrast) == 0:
        Hp = np.hstack([np.eye(p-1), -np.ones((p-1,1))])
        Hq = np.hstack([np.eye(q-1), -np.ones((q-1,1))])

        if self.hypothesis == "INTERACTION":
            H = np.kron(Hp, Hq)
        elif self.hypothesis == "PRIMARY":
            H = Hp
        elif self.hypothesis == "SECONDARY":
            H = Hq
    else:
        H = contrast

    if self.hypothesis == "INTERACTION":
        contrast_final = H @ np.kron(Ap, Aq)
    elif self.hypothesis == "PRIMARY":
         assert v is not None, "v must be defined before using reshape"
         contrast_final = H @ np.kron(Ap, v.reshape(1, -1))
    elif self.hypothesis == "SECONDARY":
         assert u is not None, "v must be defined before using reshape"
         contrast_final = H @ np.kron(u.reshape(1, -1), Aq)
    else:
        contrast_final = contrast

    pure_data = yy[:,1:]
    A = yy[:,0]

    pvalue, stat = run_onewayBF(self, method, pure_data, contrast_final, c, self.rng, indicator_a=A)

    return pvalue, stat



from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("f-anova-py")
except PackageNotFoundError:
    pass

from .fanova import (
    functionalANOVA,
    ANOVALabels,
    ANOVAUnits,
    ANOVATables,
    ANOVAMethods,
    ANOVAGroups,
)

__all__ = [
    "functionalANOVA",
    "ANOVALabels",
    "ANOVAUnits",
    "ANOVATables",
    "ANOVAMethods",
    "ANOVAGroups",
]


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import os
import math
from fanova import utils
from typing import Union, Optional, Any, Tuple


class PlotOptions:
    # Class-level constant of valid legend locations
    _VALID_LEGEND_LOCATIONS = [
        "upper right",
        "upper left",
        "lower left",
        "lower right",
        "right",
        "center left",
        "center right",
        "lower center",
        "upper center",
        "center",
        "best",
    ]

    def __init__(
        self,
        observation_size_label: bool=True,
        x_scale: str = "linear",
        y_scale: str = "linear",
        color_scale: str = "linear",
        data_transparency: float = 0.1,
        legend_transparency: float = 0.3333,
        data_line_width: float = 1.75,
        mean_line_width: float = 5,
        font_size: int = 18,
        title_labels: str = '',
        save_path: str = "",
        legend_location: str = "best",
        num_columns: int = 1,
        legend_title: str = "",
        new_colors: Optional[np.ndarray] = None,
        position: Tuple[int, int, int, int] = (90, 90, 1400, 800),
    ):

        self.observation_size_label = observation_size_label
        self.x_scale = x_scale
        self.y_scale = y_scale
        self.color_scale = color_scale
        self.data_transparency = data_transparency
        self.legend_transparency = legend_transparency
        self.data_line_width = data_line_width
        self.mean_line_width = mean_line_width
        self.font_size = font_size
        self.title_labels = title_labels
        self.save_labels = title_labels
        self.save_path = save_path
        self.legend_location = legend_location
        self.num_columns = num_columns
        self.legend_title = legend_title
        self.new_colors = new_colors
        self.position = position

        self._validate()

    def _validate(self):
        # scales
        valid_scales = ("linear", "log", "symlog", "logit")
        if self.x_scale not in valid_scales:
            raise ValueError(f"x_scale must be one of {valid_scales}, got {self.x_scale!r}")
        if self.y_scale not in valid_scales:
            raise ValueError(f"y_scale must be one of {valid_scales}, got {self.y_scale!r}")

        valid_color_scales = ("linear", "log")
        if self.color_scale not in valid_color_scales:
            raise ValueError(f"color_scale must be one of {valid_color_scales}, got {self.color_scale!r}")

        # alpha-like params
        if not (0.0 <= self.data_transparency <= 1.0):
            raise ValueError(f"data_transparency must be in [0, 1], got {self.data_transparency}")
        if not (0.0 <= self.legend_transparency <= 1.0):
            raise ValueError(f"legend_transparency must be in [0, 1], got {self.legend_transparency}")

        # legend options
        if not isinstance(self.legend_title, str):
            raise TypeError(f"legend_title must be a string, got {type(self.legend_title).__name__}")
        if self.legend_location not in self._VALID_LEGEND_LOCATIONS:
            raise ValueError(
                f"legend_location must be one of {self._VALID_LEGEND_LOCATIONS}, got {self.legend_location!r}"
            )

        # layout options
        if not isinstance(self.num_columns, int) or self.num_columns < 1:
            raise ValueError("num_columns must be an integer >= 1")
        if not (isinstance(self.position, tuple) and len(self.position) == 4 and all(isinstance(v, int) for v in self.position)):
            raise ValueError("position must be a tuple of 4 integers, e.g. (x, y, width, height)")

        # save_path: if provided, it must be an existing directory
        if isinstance(self.save_path, str) and self.save_path != "":
            normalized = os.path.abspath(os.path.expanduser(self.save_path))
            if not os.path.isdir(normalized):
                raise ValueError(f"save_path must be an existing directory, got {self.save_path!r}")
            self.save_path = normalized  # normalize

    def update_from_dict(self, updates: dict) -> None:
        """
        Update PlotOptions from a dictionary.
        Only applies changes if value is not None and different.
        """
        for key, value in updates.items():
            if value is not None and hasattr(self, key):
                if getattr(self, key) != value:
                    setattr(self, key, value)
        self._validate()

def plot_means(self, plot_type):

    plot_type = plot_type.upper()
    n_labels = []
    the_labels  = []
    fig_label = 'Group'

    if self._groups.subgroup_indicator is None:  # One Way Analysis
        the_labels = self._labels.group
        n_labels = self.n_i
        plot_type = 'DEFAULT'
        fig_label = 'Primary Factor'
        
    else: # Two Way Analysis
        if plot_type == ['PRIMARY']:
            plot_type = 'PRIMARY'
            the_labels = self._labels.primary
            n_labels = self.n_i
            fig_label = 'Primary Factor'

        elif plot_type == "SECONDARY":
            the_labels = self._labels.secondary
            n_labels = np.zeros(self._groups.B)
            fig_label = 'Secondary Factor'
            for k in range(self._groups.A):
                for kk in range(self._groups.B):
                    n_labels[kk] += self.n_ii[k][kk]
                    
        elif plot_type in ["INTERACTION", "FAMILY"]:
            plot_type = "INTERACTION"
            the_labels = utils.generate_two_way_comb(self)
            n_labels = np.concatenate([item for item in self.n_ii])
            fig_label = 'Primary & Secondary Factors'


    if self.plottingOptions.observation_size_label:
        if self._labels.generic_group:
            the_data_labels = [f": Group Data ({int(n)})" for n in n_labels]
        else:
            the_data_labels = [f": Data ({int(n)})" for n in n_labels]
    else:
        the_data_labels = [''] * len(n_labels)

    if self._labels.generic_group:
        mean_group_labels = [f"{label}: Group Mean" for label in the_labels]
        data_group_labels = [f"{the_labels[i]}{the_data_labels[i]}" for i in range(len(the_labels))]
    else:
        mean_group_labels = [f"{label}: Mean" for label in the_labels]
        data_group_labels = [f"{the_labels[i]}{the_data_labels[i]}" for i in range(len(the_labels))]


    # fast patch for figure scale errors
    fig_width = max(self.plottingOptions.position[2] / 100, 8)
    fig_height = max(self.plottingOptions.position[3] / 100, 6)

    fig, ax = plt.subplots(figsize=(fig_width, fig_height))
    
    mgr = fig.canvas.manager

    if mgr is not None:
        mgr.set_window_title(f'{fig_label} Means Visualized') 

    match plot_type:
        case 'DEFAULT':
            mean_groups = [np.mean(data, axis=1) for data in self.data]

            if self.plottingOptions.new_colors is None:
                cmap = plt.get_cmap("tab10" if self._groups.k < 7 else "turbo")
                color_list = cmap(np.linspace(0, 1, self._groups.k))
            else:
                assert self.plottingOptions.new_colors.shape[1] == 4, 'Color order matrix must have 4 columns representing RGB values from 0 to 1 with the last column being alpha the transparency'
                assert self.plottingOptions.new_colors.shape[0] >= self._groups.k, f'Color order matrix must have at least {self._groups.k} rows representing all the groups'
                color_list = self.plottingOptions.new_colors

            plot_reals = []
            legend_example_lines = []

            for k in range(self._groups.k):
                lines = ax.plot(self.d_grid, self.data[k], color=color_list[k], linewidth=self.plottingOptions.data_line_width, alpha=self.plottingOptions.data_transparency)
                plot_reals.append(lines)

                legend_line = Line2D([0], [0], color=color_list[k], linewidth=self.plottingOptions.data_line_width, alpha=self.plottingOptions.legend_transparency)
                legend_example_lines.append(legend_line)

            plot_means = []

            for k in range(self._groups.k):
                line, = ax.plot(self.d_grid, mean_groups[k], color=color_list[k],
                            linewidth=self.plottingOptions.mean_line_width, linestyle='--')
                plot_means.append(line)

            ax.set_title(f'F-ANOVA Group Means and Realizations {self.plottingOptions.title_labels}', fontsize=self.plottingOptions.font_size)
            lg = ax.legend(plot_means + legend_example_lines, mean_group_labels + data_group_labels, loc=self.plottingOptions.legend_location, ncol=self.plottingOptions.num_columns)

            if self.plottingOptions.legend_title:
                lg.set_title(self.plottingOptions.legend_title)

        case 'PRIMARY':
            mean_groups = [np.mean(data, axis=1) for data in self.data]

            if self.plottingOptions.new_colors is None:
                cmap = plt.get_cmap("tab10" if self._groups.k < 7 else "turbo")
                color_list = cmap(np.linspace(0, 1, self._groups.k))
            else:
                color_list = self.plottingOptions.new_colors

            plot_reals = []
            legend_example_lines = []

            for k in range(self._groups.k):
                lines = ax.plot(self.d_grid, self.data[k], color=color_list[k], linewidth=self.plottingOptions.data_line_width, alpha=self.plottingOptions.data_transparency)
                plot_reals.append(lines)

                legend_line = Line2D([0], [0], color=color_list[k], linewidth=self.plottingOptions.data_line_width, alpha=self.plottingOptions.legend_transparency)
                legend_example_lines.append(legend_line)

            plot_means = []

            for k in range(self._groups.k):
                line, = ax.plot(self.d_grid, mean_groups[k], color=color_list[k], linewidth=self.plottingOptions.mean_line_width, linestyle='--')
                plot_means.append(line)

            ax.set_title(f'F-ANOVA Primary Factor Means and Realizations {self.plottingOptions.title_labels}', fontsize=self.plottingOptions.font_size)
            lg = ax.legend(plot_means + legend_example_lines, mean_group_labels + data_group_labels, loc=self.plottingOptions.legend_location, ncol=self.plottingOptions.num_columns)

            if not self.plottingOptions.legend_title:
                lg.set_title("Primary Factor Levels")
            else:
                lg.set_title(self.plottingOptions.legend_title)

        case 'SECONDARY':
            yy = np.concatenate([data.T for data in self.data], axis=0)

            bflag = self._groups.subgroup_indicator
            bflag0 = np.unique(bflag)
            n_secondary = len(bflag0)

            if self.plottingOptions.new_colors is None:
                cmap = plt.get_cmap("tab10" if n_secondary < 7 else "turbo")
                color_list = cmap(np.linspace(0, 1,n_secondary))
            else:
                color_list = self.plottingOptions.new_colors

            sub_data = []
            mean_groups = []

            for k in range(n_secondary):
                flag = bflag == bflag0[k]
                sub_data_k = yy[flag, :].T
                sub_data.append(sub_data_k)
                mean_groups.append(np.mean(sub_data_k, axis=1))

            plot_reals = []
            legend_example_lines = []

            for k in range(n_secondary):
                lines = ax.plot(self.d_grid, sub_data[k], color=color_list[k], linewidth=self.plottingOptions.data_line_width, alpha=self.plottingOptions.data_transparency)
                plot_reals.append(lines)

                legend_line = Line2D([0], [0], color=color_list[k], linewidth=self.plottingOptions.data_line_width, alpha=self.plottingOptions.legend_transparency)
                legend_example_lines.append(legend_line)

            plot_means = []

            for k in range(n_secondary):
                line, = ax.plot(self.d_grid, mean_groups[k], color=color_list[k],
                            linewidth=self.plottingOptions.mean_line_width, linestyle='--')
                plot_means.append(line)

            ax.set_title(f'F-ANOVA Secondary Factor Means and Realizations {self.plottingOptions.title_labels}', fontsize=self.plottingOptions.font_size)
            lg = ax.legend(plot_means + legend_example_lines, mean_group_labels + data_group_labels, loc=self.plottingOptions.legend_location, ncol=self.plottingOptions.num_columns)

            if not self.plottingOptions.legend_title:
                lg.set_title("Secondary Factor Levels")
            else:
                lg.set_title(self.plottingOptions.legend_title)

        case 'INTERACTION':
            yy = np.concatenate([data.T for data in self.data], axis=0)

            aflag = np.repeat(np.arange(1, self._groups.A + 1), self.n_i)
            bflag = self._groups.subgroup_indicator
            aflag0 = np.unique(aflag)
            bflag0 = np.unique(bflag)
            p, q = len(aflag0), len(bflag0)
            ab = p * q

            combinations = utils.generate_two_way_comb(self)
            mean_labels = [f"{comb}: Mean" for comb in combinations]

            if self.plottingOptions.observation_size_label:
                data_labels = [f"{combinations[k]}: Data ({int(n_labels[k])})" for k in range(len(combinations))]
            else:
                data_labels = [f"{comb}: Data" for comb in combinations]

            if self.plottingOptions.new_colors is None:
                cmap = plt.get_cmap("tab10" if ab < 7 else "turbo")
                color_list = cmap(np.linspace(0, 1, ab))
            else:
                color_list = self.plottingOptions.new_colors

            sub_data = []
            mean_groups = []
            counter = 0

            for i in aflag0:
                for j in bflag0:
                    ijflag = (aflag == i) & (bflag == j)
                    yyi = yy[ijflag, :]
                    sub_data.append(yyi.T)
                    mean_groups.append(np.mean(yyi, axis=0))
                    counter += 1

            plot_reals = []
            legend_example_lines = []

            for k in range(ab):
                lines = ax.plot(self.d_grid, sub_data[k], color=color_list[k], linewidth=self.plottingOptions.data_line_width, alpha=self.plottingOptions.data_transparency)
                plot_reals.append(lines)

                legend_line = Line2D([0], [0], color=color_list[k],linewidth=self.plottingOptions.data_line_width, alpha=self.plottingOptions.legend_transparency)
                legend_example_lines.append(legend_line)

            plot_means = []

            for k in range(ab):
                line, = ax.plot(self.d_grid, mean_groups[k], color=color_list[k], linewidth=self.plottingOptions.mean_line_width, linestyle='--')
                plot_means.append(line)

            ax.set_title(f'F-ANOVA Primary and Secondary Factor Combinatorial Means and Realizations\n{self.plottingOptions.title_labels}', fontsize=self.plottingOptions.font_size)
            lg = ax.legend(plot_means + legend_example_lines, mean_labels + data_labels, loc=self.plottingOptions.legend_location, ncol=self.plottingOptions.num_columns)

            if not self.plottingOptions.legend_title:
                lg.set_title("TwoWay Factor Levels")
            else:
                lg.set_title(self.plottingOptions.legend_title)


    ax.set_xscale(self.plottingOptions.x_scale)
    ax.set_yscale(self.plottingOptions.y_scale)


    ax.set_xlabel(auto_make_labels(self._units.domain, self._labels.domain), fontsize=self.plottingOptions.font_size)
    ax.set_ylabel(auto_make_labels(self._units.response, self._labels.response), fontsize=self.plottingOptions.font_size)

    ax.tick_params(labelsize=self.plottingOptions.font_size)

    if self.plottingOptions.save_path:

        save_filename = f"GroupMeans_{self.plottingOptions.save_labels}.png"

        fig.savefig(os.path.join(self.plottingOptions.save_path, save_filename))
        plt.close(fig)
    else:
        plt.show(block=False)

    return fig, ax

def auto_make_labels(units, quanity_label):
    if quanity_label is not None:
        axis_label = f'{quanity_label}'
    else:
        axis_label = ''

    if units is not None:
        if axis_label == '':
            axis_label = f'({units})'
        else:
            axis_label += ' ' + f'({units})'

    return axis_label

def _make_covariances(data, k, domain_points):
    eta_i = np.zeros((domain_points, k))
    v_hat_i = [None] * k

    gamma_hat_i = np.zeros((domain_points, domain_points, k))
    n_ii = np.zeros(k)

    for kk in range(k):
        n_i = data[kk].shape[1]
        n_ii[kk] = n_i
        eta_i[:, kk] = np.mean(data[kk], axis=1)
        zero_mean_data_k_subset = data[kk] - eta_i[:, kk:kk+1]
        v_hat_i[kk] = zero_mean_data_k_subset
        gamma_hat_i[:, :, kk] = (1 / (n_i - 1)) * (zero_mean_data_k_subset @ zero_mean_data_k_subset.T)

    N = int(np.sum(n_ii))
    pooled_covar_terms = np.zeros((domain_points, domain_points, k))

    for kk in range(k):
        pooled_covar_terms[:, :, kk] = (n_ii[kk] - 1) * gamma_hat_i[:, :, kk]
    pooled_covar = np.sum(pooled_covar_terms, axis=2) / (N - k)

    return gamma_hat_i, pooled_covar, n_ii

def _grid_generator(k):
    if k <= 3:
        n = k
        m = 1
    elif k == 4:
        n = 2
        m = 2
    elif k == 5:
        n = 2
        m = 3
    elif k == 6:
        n = 2
        m = 3
    elif k == 7:
        n = 3
        m = 3
    elif k == 8:
        n = 3
        m = 3
    else:
        # n = math.ceil(math.sqrt(k))
        # m = math.ceil(k / n)
        n = np.ceil(math.sqrt(k))
        m = np.ceil(k / n)
    return int(m), int(n)

def plot_covariances(self, plot_type):
    # Default Vars
    K = None
    gamma_hat_i = None 
    pooled_covar = None
    display_label = []
    n_ii = []
    
    plot_type = plot_type.upper()
    fig_label = 'Group'
    temp_label = ''

    if self._groups.subgroup_indicator is None:  # OneWay
        assert plot_type == 'FAMILY', 'TwoWay plotting options require a subgroup_indicator argument'

        plot_type = 'DEFAULT'
        if self._labels.generic_group:
            display_label = [f"Group {i+1}" for i in range(self._groups.k)]
        else:
            display_label = self._labels.group
    else:
        if plot_type == 'PRIMARY':
            plot_type = 'PRIMARY'
            fig_label = 'Primary Factor'
            display_label = self._labels.primary

            if self._labels.generic_group:
                display_label = [f"Group {label}" for label in display_label]

        elif plot_type == 'SECONDARY':
            fig_label = 'Secondary Factor'
            display_label = self._labels.secondary

            if self._labels.generic_group:
                display_label = [f"Group {label}" for label in display_label]

        elif plot_type in ['INTERACTION', "FAMILY"]:
            plot_type = 'INTERACTION'
            fig_label = 'Primary & Secondary Factors'
            combinations = utils.generate_two_way_comb(self)
            display_label = combinations

    if self.plottingOptions.title_labels:
        title_labels_str = str(self.plottingOptions.title_labels)
        save_labels = str(self.plottingOptions.title_labels)
    else:
        title_labels_str = ''
        save_labels = ''

    if not self._units.response:
        color_bar_label = r'$(\mathrm{Response})^2$'
    else:
        color_bar_label = rf'$({self._units.response})^{{2}}$'

    if plot_type in  ['DEFAULT', 'PRIMARY']:
        gamma_hat_i, pooled_covar, n_ii = _make_covariances(self.data, self._groups.k, self.n_domain_points)
        K = self._groups.k

    elif plot_type == 'SECONDARY':
        yy = np.concatenate([data.T for data in self.data], axis=0)

        bflag = self._groups.subgroup_indicator
        bflag0 = np.unique(bflag)

        n_secondary = len(bflag0)
        sub_data = []

        for k in range(n_secondary):
            flag = bflag == bflag0[k]
            sub_data.append(yy[flag, :].T)

        gamma_hat_i, pooled_covar, n_ii = _make_covariances(sub_data, n_secondary, self.n_domain_points)
        K = n_secondary

    elif plot_type == 'INTERACTION':
        yy = np.concatenate([data.T for data in self.data], axis=0)
        aflag = np.repeat(np.arange(1, self._groups.A + 1), self.n_i)
        bflag = self._groups.subgroup_indicator

        aflag0 = np.unique(aflag)
        bflag0 = np.unique(bflag)

        p, q = len(aflag0), len(bflag0)
        ab = p * q
        self._groups.AB = ab
        sub_data = []
        counter = 0

        for i in aflag0:
            for j in bflag0:
                ijflag = (aflag == i) & (bflag == j)
                yyi = yy[ijflag, :]
                sub_data.append(yyi.T)
                counter += 1

        gamma_hat_i, pooled_covar, n_ii = _make_covariances(sub_data, ab, self.n_domain_points)
        K = ab

    assert K is not None, 'K is None and it should have been assigned'
    assert gamma_hat_i is not None, 'covariance wasnt calculated and it should have been'
    assert pooled_covar is not None, 'pooled covariance wasnt calculated and it should have been'

    
    fig_width = max(self.plottingOptions.position[2] / 100, 8)  # minimum 8 inches
    fig_height = max(self.plottingOptions.position[3] / 100, 6)  # minimum 6 inches

    fig = plt.figure(figsize=(fig_width, fig_height))
    mgr = fig.canvas.manager

    if mgr is not None:
        mgr.set_window_title(f'{fig_label} Covariances Visualized')

    n_plots = K + 1
    rows, cols = _grid_generator(n_plots)

    extent = (float(self.d_grid[0]),float(self.d_grid[-1]),float(self.d_grid[0]),float(self.d_grid[-1]),)

    for ii in range(K):
        ax = plt.subplot(rows, cols, ii + 1)
        cmin = np.min(gamma_hat_i[:, :, ii])
        cmax = np.max(gamma_hat_i[:, :, ii])

        im = ax.imshow(gamma_hat_i[:, :, ii], extent=extent,
                      origin='lower', aspect='equal', vmin=cmin, vmax=cmax)
        ax.set_title(f"{display_label[ii]} | n = {int(n_ii[ii])}")
        cbar = plt.colorbar(im, ax=ax)
        cbar.set_label(color_bar_label)
        
        ax.set_xticks(ax.get_yticks())

        if self._units.domain:
            temp_label = f"({self._units.domain})"

        ax.set_xlabel(temp_label)
        ax.set_ylabel(temp_label)

        if not self.plottingOptions.x_scale:
            ax.set_xscale('linear')
        else:
            ax.set_xscale(self.plottingOptions.x_scale)

        if not self.plottingOptions.y_scale:
            ax.set_yscale('linear')
        else:
            ax.set_yscale(self.plottingOptions.y_scale)

    ax = plt.subplot(rows, cols, K + 1)
    cmin = np.min(pooled_covar)
    cmax = np.max(pooled_covar)

    im = ax.imshow(pooled_covar, extent=extent,
                  origin='lower', aspect='equal', vmin=cmin, vmax=cmax)

    ax.set_title(f"Pooled | n = {self.N}")
    ax.set_xlabel(temp_label)
    ax.set_ylabel(temp_label)
    ax.set_xticks(ax.get_yticks())
    
    cbar = plt.colorbar(im, ax=ax)
    cbar.set_label(color_bar_label)


    plt.suptitle(f'{fig_label} Covariances {title_labels_str}', fontsize=10)

    for ax in fig.get_axes():
        ax.tick_params(labelsize=self.plottingOptions.font_size)
        for text in ax.get_xticklabels() + ax.get_yticklabels():
            text.set_fontsize(self.plottingOptions.font_size)

    plt.tight_layout()

    if not self.plottingOptions.x_scale:
        plt.gca().set_xscale('linear')
    else:
        plt.gca().set_xscale(self.plottingOptions.x_scale)

    if not self.plottingOptions.y_scale:
        plt.gca().set_yscale('linear')
    else:
        plt.gca().set_yscale(self.plottingOptions.y_scale)

    if self.plottingOptions.color_scale == 'log':
        if hasattr(plt.gca(), 'set_norm'):
            from matplotlib.colors import LogNorm
            plt.gca().images[0].set_norm(LogNorm())

    if self.plottingOptions.save_path and os.path.isdir(self.plottingOptions.save_path):
        save_filename = f"Covariance_{save_labels}.png"
        fig.savefig(os.path.join(self.plottingOptions.save_path, save_filename))
        plt.close(fig)
    else:
        plt.show(block=False)

    return fig

# TODO: Some option or ability to save these plots
def plot_test_stats(self, p_value:float | np.floating, null_dist:np.ndarray, test_stat:float | np.floating,
                    test_name:str, scedasticity:str,  k:int, N:int|None = None):

    p_value = float(p_value)

    if p_value <= self.alpha:
        line_label = f'{test_name} Statistic P-Value: p={p_value: 0.2f} <= {self.alpha}'
        verdict_label = 'Verdict: Reject $H_0$'

    else:
        line_label = f'{test_name} Statistic P-Value: p={p_value: 0.2f} > {self.alpha}'
        verdict_label = 'Verdict: Fail to Reject $H_0$'

    fig, ax = plt.subplots(figsize=(10, 8))

    ax.set_title(f'Null Distribution Plot for {test_name}')
    n, bins, patches = ax.hist(null_dist, bins=100, density=True, alpha=0.7, label='Null Distribution', color='grey', edgecolor='black')

    ax.axvline(float(test_stat), color='blue', linestyle='--', linewidth=1.5, label=line_label)

    ax.text(
        float(test_stat),                          # x: aligned with the line
        ax.get_ylim()[1] * 0.5,             # y: middle of the y-axis
        f'p = {p_value:.2f}',
        color='blue',
        rotation=90,                        # make it vertical
        va='center', ha='right',            # vertical and horizontal alignment
        fontsize=10,
        backgroundcolor='white'             # optional: makes text readable over bars
    )

    if (test_stat > np.max(null_dist) * 1e2) or (test_stat < np.min(null_dist) * 1e2):
        ax.set_xscale('log')
        ax.minorticks_on()

    crit_value = np.quantile(null_dist, 1 - self.alpha)

    x_max = max(np.max(null_dist), test_stat) * 1.2  # ensure room for shading
    ax.set_xlim(right=x_max)

    ax.axvspan(crit_value, x_max, color='red', alpha=0.3,
            label='Region for Statistical Significance (Reject H₀)')


    ax.axvline(crit_value, color='black', linestyle='-', linewidth=1.5, label='Beginning of Critical Value Region')

    # Add text label ON the shaded region
    ax.text(
        crit_value * 1.5,                # X position: move into the shaded region
        ax.get_ylim()[1] * 0.9,          # Y position: near top
        'Critical Region\n(Reject $H_0$)',
        color='darkred',
        fontsize=10,
        ha='left',
        va='top',
        bbox=dict(facecolor='white', edgecolor='none', alpha=0.7)  # optional background
)

    title_label = ''
    if 'f' in test_name.lower():
        assert N is not None and isinstance(N, int), "'N' must be provided for F-type mixture labeling"

        null_dist_label = fr'Simulated $F_{{(k-1),(n-k)}} = F_{{({k - 1}),({N - k})}}$-type Mixture Null Distribution'
        if self.hypothesis == "FAMILY":
            title_label = f'One-Way({scedasticity}), Family, Functional ANOVA: F-type test'
        elif self.hypothesis == "PAIRWISE":
            title_label = (f"One-Way({scedasticity}), Pairwise ({self._labels.hypothesis})\n" r"Functional ANOVA: F-type test")
    else:
        null_dist_label = fr'Simulated $\chi^2_{{(k-1)}} = \chi^2_{{{k - 1}}}$-type Mixture Null Distribution'
        if self.hypothesis == "FAMILY":
            title_label = fr'One-Way ({scedasticity}), Family, Functional ANOVA: Squared $L^2$ Norm test'
        elif self.hypothesis== "PAIRWISE":
            title_label = (f"One-Way ({scedasticity}), Pairwise ({self._labels.hypothesis})\n"
                           r"Functional ANOVA: F-type test")


    ax.set_ylabel('PDF', fontsize=14)
    ax.set_xlabel(null_dist_label, fontsize=14)
    ax.set_title(f'{title_label} \n{verdict_label}', fontsize=14)

    ax.legend(
    loc='upper center',
    bbox_to_anchor=(0.5, -0.15),  # Move it further down
    ncol=2,
    fontsize=12,
    frameon=False  # Optional: removes legend border)
    )

    ax.tick_params(labelsize=14)

    plt.tight_layout()
    plt.show(block=False)

    return
